subroutine cubspl4 (n,tau,c1,c2,c3,c4,ibcbeg,ibcend,error)
  use greg_interfaces, except_this=>cubspl4
  use gbl_message
  !---------------------------------------------------------------------
  ! @ private
  ! GREG	Internal routine
  !
  !	Piecewise cubic spline interpolants computation; adapted from
  !	'A practical guide to SPLINES' , CARL DE BOOR ,
  !	Applied Mathematical Sciences, SPRINGER-VERLAG, VOL.27, P57-59 (1978).
  !
  ! Arguments :
  !     	N	I	Number of data points. assumed to be .ge. 2.	Input
  !	TAU	R*4(N)	Abscissae of data points, strictly monotonous	Input
  !	C1	R*4(N)	Ordinates of data points			Input
  !     	IBCBEG	I							Input
  !	IBCEND  I	Boundary condition indicators, and		Input
  !	C1	R*4(N)	First polynomial coefficient of spline	Output
  !	C2	R*4(N)	Second   --         --           --	Output/Input
  !	C3	R*4(N)	Third    --         --           --	Output
  !	C4	R*4(N)	Fourth   --         --           --	Output
  !
  ! 	C2(1), C2(N)  are boundary condition information.
  !	Specifically,
  !	IBCBEG = 0  means no boundary condition at tau(1) is given.
  !		In this case, the not-a-knot condition is used, i.e. the
  !		jump in the third derivative across TAU(2) is forced to
  !		zero, thus the first and the second cubic polynomial pieces
  !		are made to coincide.
  !        IBCBEG = 1 means that the slope at TAU(1) is made to equal
  !		C2(1), supplied by input.
  !        IBCBEG = 2 means that the second derivative at TAU(1) is
  !		made to equal C2(1), supplied by input.
  !        IBCEND = 0, 1, or 2 has analogous meaning concerning the
  !		boundary condition at TAU(N), with the additional
  !		information taken from C2(N).
  !
  !	CJ(I), J=1,...,4; I=1,...,L (= N-1) are the polynomial
  !	coefficients of the cubic interpolating spline. Precisely,
  !	in the interval    [ TAU(I), TAU(I+1) ]    the spline F
  !	is given by
  !		F(X) = C1(I)+H*(C2(I)+H*(C3(I)+H*C4(I)/3.)/2.)
  !	where H = X - TAU(I).
  !
  !	In other words, for I=1,...,N, C2(I) and C3(I) are respectively
  !	equal to the values of the first and second derivatives of
  !	the interpolating spline, and C4(I) is equal to the third
  !	derivative of the interpolating spline in the interval
  !	[ TAU(I), TAU(I+1) ]. C4(N) is meaningless and is set to 0.
  !---------------------------------------------------------------------
  integer :: n                      !
  real :: tau(n)                    !
  real :: c1(n)                     !
  real :: c2(n)                     !
  real :: c3(n)                     !
  real :: c4(n)                     !
  integer :: ibcbeg                 !
  integer :: ibcend                 !
  logical :: error                  !
  ! Local
  character(len=*), parameter :: rname='CUBSPL4'
  real :: taum1,g,dtau,divdf1,divdf3
  integer :: i,j,l,m
  !
  ! A tridiagonal linear system for the unknown slopes S(I) of
  ! F at TAU(I), I=1,...,N, is generated and then solved by gauss
  ! elimination, with S(I) ending up in C2(I), all I.
  ! C3(.) and C4(.) are used initially for temporary storage.
  !
  ! Check if N.GE.2
  !
  if (n.lt.2) then
    call greg_message(seve%e,rname,'Less than two pivots')
    error = .true.
  endif
  !
  ! Check if TAU strictly monotonous
  taum1=tau(2)
  if (tau(2).lt.tau(1)) then
    do i=3,n
      if (tau(i).ge.taum1) then
        call greg_message(seve%e,rname,'Variable is not strictly monotonous')
        error = .true.
        return
      endif
      taum1 = tau(i)
    enddo
  elseif (tau(2).gt.tau(1)) then
    do i=3,n
      if (tau(i).le.taum1) then
        call greg_message(seve%e,rname,'Variable is not strictly monotonous')
        error = .true.
        return
      endif
      taum1 = tau(i)
    enddo
  else
    call greg_message(seve%e,rname,'Variable is not strictly monotonous')
    error = .true.
    return
  endif
  !
  l = n-1
  !
  ! Compute first differences of TAU sequence and store in C3(.).
  ! Also, compute first divided difference of data and store in C4(.).
  !
  do  m=2,n
    c3(m) = tau(m) - tau(m-1)
    c4(m) = (c1(m) - c1(m-1))/c3(m)
  enddo
  !
  ! Construct first equation from the boundary condition, of the form
  !             C4(1)*S(1) + C3(1)*S(2) = C2(1)
  !
  if (ibcbeg-1)                     11,15,16
11 if (n .gt. 2)                     goto 12
  !
  ! No condition at left end and N = 2.
  c4(1) = 1.
  c3(1) = 1.
  c2(1) = 2.*c4(2)
  goto 25
  !
  ! Not-a-knot condition at left end and N .GT. 2.
12 c4(1) = c3(3)
  c3(1) = c3(2) + c3(3)
  c2(1) = ((c3(2)+2.*c3(1))*c4(2)*c3(3)+c3(2)**2*c4(3))/c3(1)
  goto 19
  !
  ! Slope prescribed at left end.
15 c4(1) = 1.
  c3(1) = 0.
  goto 18
  !
  ! Second derivative prescribed at left end.
16 c4(1) = 2.
  c3(1) = 1.
  c2(1) = 3.*c4(2) - c3(2)/2.*c2(1)
18 if (n.eq.2)  goto 25
  !
  ! If there are interior knots, generate the corresponding equations and
  ! carry out the forward pass of gauss elimination, after which the M-th
  ! equation reads    C4(M)*S(M) + C3(M)*S(M+1) = C2(M).
  !
19 continue
  do m=2,l
    g = -c3(m+1)/c4(m-1)
    c2(m) = g*c2(m-1) + 3.*(c3(m)*c4(m+1)+c3(m+1)*c4(m))
    c4(m) = g*c3(m-1) + 2.*(c3(m) + c3(m+1))
  enddo
  !
  ! Construct last equation from the second boundary condition, of the form
  !           (-G*C4(N-1))*S(N-1) + C4(N)*S(N) = C2(N)
  ! If slope is prescribed at right end, one can go directly to back
  ! substitution, since C array happens to be set up just right for it
  ! at this point.
  !
  if (ibcend-1)                     21,30,24
21 if (n .eq. 3 .and. ibcbeg .eq. 0) goto 22
  !
  ! not-a-knot and N.GE.3, and either N.GT.3 or also not-a-knot at
  ! left endpoint.
  !
  g = c3(l) + c3(n)
  c2(n) = ((c3(n)+2.*g)*c4(n)*c3(l) + c3(n)**2*(c1(l)-c1(n-2))/c3(l))/g
  g = -g/c4(l)
  c4(n) = c3(l)
  goto 29
  !
  ! Either (N=3 and not-a-knot also at left) or (N=2 and not not-a-knot
  ! at left endpoint).
  !
22 c2(n) = 2.*c4(n)
  c4(n) = 1.
  goto 28
  !
  ! Second derivative prescribed at right endpoint.
24 c2(n) = 3.*c4(n) + c3(n)/2.*c2(n)
  c4(n) = 2.
  goto 28
25 if (ibcend-1)                     26,30,24
26 if (ibcbeg .gt. 0)                goto 22
  !
  ! not-a-knot at right endpoint and at left endpoint and N=2.
  c2(n) = c4(n)
  goto 30
28 g = -1./c4(l)
  !
  ! Complete forward pass of GAUSS elimination.
29 c4(n) = g*c3(l) + c4(n)
  c2(n) = (g*c2(l) + c2(n))/c4(n)
  !
  ! Carry out back substitution
30 continue
  do j=l,1,-1
    c2(j) = (c2(j) - c3(j)*c2(j+1))/c4(j)
  enddo
  !
  ! Generate cubic coefficients in each interval, i.e., the derivatives
  ! at its left endpoint, from value and slope at its endpoints.
  !
  do i=2,n
    dtau = c3(i)
    divdf1 = (c1(i) - c1(i-1))/dtau
    divdf3 = c2(i-1) + c2(i) - 2.*divdf1
    c3(i-1) = 2.*(divdf1 - c2(i-1) - divdf3)/dtau
    c4(i-1) = (divdf3/dtau)*(6./dtau)
  enddo
  !
  ! Compute in addition C3(N). set C4(N) to 0.
  c3(n) = c3(l) + c4(l)*dtau
  c4(n) = 0.
end subroutine cubspl4
