subroutine atm_init(error)
  use gbl_message
  use gkernel_interfaces
  use atm_params
  use ast_astro
  !---------------------------------------------------------------------
  ! @ private
  ! ATM
  !
  ! One time instantiation of the ASTRO variables which are used (as
  ! input and output) by the command ASTRO\ATM. These variables and
  ! their startup values are not needed if the programmer does want
  ! to make use of the command ASTRO\ATM.
  !---------------------------------------------------------------------
  logical, intent(inout) :: error   ! Logical error flag
  ! Local
  character(len=*), parameter :: rname='ATM_INIT'
  !
  if (atm_initdone) return
  !
  call atm_sicvariables(error)
  if (error) then
    call astro_message(seve%e,rname,'Error creating ATM% variables')
    return
  endif
  !
  ! Default values of input parameters
  t0 = 273.0
  p0 = 1013.0
  water = 1.0
  airmass = 1.414
  feff = 0.95
  freqs = 110.0
  freqi = 113.0
  gim = 1.0
  trec = 60
  !
  ! Input parameters
  call sic_def_real ('TEMPERATURE',t0,0,1,.false.,error)
  if (error) return
  call sic_def_real ('TRUE_PRESSURE',p1,0,1,.true.,error)
  if (error) return
  call sic_def_real ('ZERO_PRESSURE',p0,0,1,.false.,error)
  if (error) return
  call sic_def_dble ('ALTITUDE',altitude,0,1,.true.,error)
  if (error) return
  call sic_def_real ('WATER',water,0,1,.false.,error)
  !! if (error) return
  call sic_def_real ('AIRMASS',airmass,0,1,.false.,error)
  if (error) return
  call sic_def_real ('FORWARD_EFF',feff,0,1,.false.,error)
  !! if (error) return
  call sic_def_real ('FREQ_SIG',freqs,0,1,.false.,error)
  if (error) return
  call sic_def_real ('FREQ_IMA',freqi,0,1,.false.,error)
  if (error) return
  call sic_def_real ('GAIN_IMAGE',gim,0,1,.false.,error)
  !! if (error) return
  call sic_def_real ('TREC',trec,0,1,.false.,error)
  if (error) return
  !
  ! Output parameters
  call sic_def_real ('TSYS',tsys,0,1,.true.,error)
  !! if (error) return
  call sic_def_real ('TAU_O2',tauox,0,1,.true.,error)
  if (error) return
  call sic_def_real ('TAU_H2O',tauw,0,1,.true.,error)
  if (error) return
  call sic_def_real ('TAU_TOT',taut,0,1,.true.,error)
  if (error) return
  call sic_def_real ('EMIS_SIG',temis,0,1,.true.,error)
  if (error) return
  call sic_def_real ('EMIS_IMA',temii,0,1,.true.,error)
  if (error) return
  call sic_def_real ('PATH_SIG',paths,0,1,.true.,error)
  if (error) return
  call sic_def_real ('PATH_IMA',pathi,0,1,.true.,error)
  if (error) return
  call sic_def_real ('ATM_SIG',tatms,0,1,.true.,error)
  if (error) return
  call sic_def_real ('ATM_IMA',tatmi,0,1,.true.,error)
  if (error) return
  call sic_def_real ('TANT',tant,0,1,.true.,error)
  if (error) return
  !
  call astro_message(seve%d,rname,'ATM-ASTRO interface initialized')
  atm_initdone = .true.
  !
end subroutine atm_init
!
subroutine astro_atm(line,error)
  use gildas_def
  use gkernel_interfaces
  use atm_params
  use ast_astro
  !---------------------------------------------------------------------
  ! @ private
  ! ATM
  !
  ! Purpose: compute atmospheric properties
  !     from a full set of SIC variables
  !---------------------------------------------------------------------
  character(len=*), intent(in)    :: line   ! Input command line
  logical,          intent(inout) :: error  ! Logical error flag
  ! Local
  integer(kind=4) :: ier,nfile,nf
  real(kind=4) :: frequ,fmin,fmax
  character(len=filename_length) :: file
  integer(kind=4), parameter :: mact=3
  integer(kind=4) :: nact,nname
  character(len=16) :: actions(mact),action,name
  logical, save :: first
  ! Data
  data actions/'INTERPOLATE','MAKE','SAVE'/
  data first /.true./
  !
  ! Optionally print ATM status before the actions below
  if (sic_present(1,0))  call astro_atm_print(error)
  !
  h0 = altitude
  if (.not.sic_present(0,1)) then
    !
    p1 = p0*2.0**(-h0/5.5)     ! Pressure at altitude H0
    call atm_atmosp(t0,p1,h0)
    !
    frequ = freqi
    call atm_transm(water,airmass,frequ,temii,tatmi,tauox,tauw,taut,ier)
    call atm_path(water,airmass,frequ,pathi,ier)
    pathi = pathi*10           ! -> mm
    !
    frequ = freqs
    call atm_transm(water,airmass,frequ,temis,tatms,tauox,tauw,taut,ier)
    call atm_path(water,airmass,frequ,paths,ier)
    paths = paths*10           ! -> mm
    !
    tant = (gim * (feff*temii+(1.0-feff)*t0 + trec) +  &
           (feff*temis+(1.0-feff)*t0 + trec) ) / (1.0+gim)
    tsys = exp(taut*airmass) * (gim * (feff*temii+(1.0-feff)*t0 + trec) +  &
           (feff*temis+(1.0-feff)*t0 + trec) ) / feff
    !
  else
    ! Read binary file once.
    if (first) then
      call atm_i(error)
      first = .false.
    endif
    ! Retrieve and analyze first argument
    call sic_ke(line,0,1,name,nname,.true.,error)
    if (error) return
    call sic_ambigs('ATMOSPHERE',name(1:nname),action,nact,actions,mact,error)
    if (error) return
    !
    if (action.eq.'INTERPOLATE') then
      ! Interpolate from file:
      call atm_atmosp_i(t0,p1,h0)
      !
      frequ = freqi
      call atm_transm_i(water,airmass,frequ,temii,tatmi,tauox,tauw,taut,ier)
      call atm_path_i(water,airmass,frequ,pathi,ier)
      pathi = pathi*10         ! -> mm
      !
      frequ = freqs
      call atm_transm_i(water,airmass,frequ,temis,tatms,tauox,tauw,taut,ier)
      call atm_path_i(water,airmass,frequ,paths,ier)
      paths = paths*10         ! -> mm
      !
      tant = (gim * (feff*temii+(1.0-feff)*t0 + trec) +  &
             (feff*temis+(1.0-feff)*t0 + trec) ) / (1.0+gim)
      tsys = exp(taut*airmass) * (gim * (feff*temii+(1.0-feff)*t0 + trec) +  &
             (feff*temis+(1.0-feff)*t0 + trec) ) / feff
    !
    elseif (action.eq.'MAKE') then
      ! Make a file: ATM Make filename nf fmin fmax
      file = ' '
      call sic_ch(line,0,2,file,nfile,.true.,error)
      if (error) return
      nf = 200
      fmin = 75.
      fmax = 300.
      call sic_i4(line,0,3,nf,.false.,error)
      if (error) return
      call sic_r4(line,0,4,fmin,.false.,error)
      if (error) return
      call sic_r4(line,0,5,fmax,.false.,error)
      if (error) return
      call atmos_i_table(.true.,file,nfile,nf,fmin,fmax,h0,error)
    !
    elseif (action.eq.'SAVE') then
      ! Save on file previously loaded table: ATM Save filename
      ! This one may be used to save with the current machine byte order a
      ! table previously loaded whose byte order was different.
      file = ' '
      call sic_ch(line,0,2,file,nfile,.true.,error)
      if (error) return
      nf = 200
      fmin = 75.
      fmax = 300.
      call atmos_i_table(.false.,file,nfile,nf,fmin,fmax,h0,error)
    endif
  endif
  !
end subroutine astro_atm
!
subroutine astro_atm_print(error)
  use gbl_message
  use atm_params
  use ast_astro
  !---------------------------------------------------------------------
  ! @ private
  ! ATM
  !
  ! Support routine for command:
  !   ASTRO\ATM /PRINT
  !---------------------------------------------------------------------
  logical, intent(inout) :: error   ! Logical error flag
  ! Local
  character(len=*), parameter :: rname='ATM_PRINT'
  character(len=message_length) :: mess
  !
  call astro_message(seve%r,rname,'Current status of ATM is')
  !
  ! Print ATM internal parameters
  call atm_print(error)
  !
  ! Print ASTRO_ATM internal parameters
  write(mess,101) 'TEMPERATURE',t0,'[K]','Ground temperature'
  call astro_message(seve%r,rname,mess)
  !
  write(mess,101) 'ZERO_PRESSURE',p0,'[hPa]','Pressure at sea level'
  call astro_message(seve%r,rname,mess)
  !
  write(mess,101) 'ALTITUDE',altitude,'[km]','Altitude of the observatory'
  call astro_message(seve%r,rname,mess)
  !
  write(mess,101) 'WATER',water,'[mm]','Precipitable water vapor'
  call astro_message(seve%r,rname,mess)
  !
  write(mess,101) 'AIRMASS',airmass,'[]','Number of airmasses'
  call astro_message(seve%r,rname,mess)
  !
  write(mess,101) 'FORWARD_EFF',feff,'[]','Forward efficiency'
  call astro_message(seve%r,rname,mess)
  !
  write(mess,101) 'FREQ_SIG',freqs,'[GHz]','Signal frequency'
  call astro_message(seve%r,rname,mess)
  !
  write(mess,101) 'FREQ_IMA',freqi,'[GHz]','Image frequency'
  call astro_message(seve%r,rname,mess)
  !
  write(mess,101) 'GAIN_IMAGE',gim,'[]','Gain image'
  call astro_message(seve%r,rname,mess)
  !
  write(mess,101) 'TREC',trec,'[K]','Receiver temperature'
  call astro_message(seve%r,rname,mess)
  !
101 format(A,T23,F12.3,T36,A6,T43,A)
  !
end subroutine astro_atm_print
