subroutine wheader_virtual(kx,head,user,assoc,error)
  use gildas_def
  use gbl_message
  use class_dependencies_interfaces
  use class_interfaces, except_this=>wheader_virtual
  use class_types
  use class_virtual
  !---------------------------------------------------------------------
  ! @ private
  !  Save the input header and user section into the virtual buffer
  !---------------------------------------------------------------------
  integer(kind=entry_length), intent(in)    :: kx     ! Entry number
  type(header),               intent(in)    :: head   ! The header
  type(class_user_t),         intent(in)    :: user   ! The user section
  type(class_assoc_t),        intent(in)    :: assoc  ! The Associated Arrays section
  logical,                    intent(inout) :: error  ! Error flag
  ! Local
  character(len=*), parameter :: rname='WHEADER_VIRTUAL'
  !
  ! Allocate memory for CLASS header buffer at first call
  if (.not.allocated(header_content)) then
    call allocate_memory_header(error)
    if (error) then
      ! Could not allocate memory for header contents. Turn off the
      ! virtual buffer and leave
      error = .false.  ! Not an error
      call class_message(seve%w,rname,  &
        'Memory allocation failed, virtual buffer turned OFF')
      call setup_virtual_off(error)
      return
    endif
  endif
  !
  ! Free next available slot if it was used
  if (header_xnum(next_header).ne.0) then
    got_header(header_xnum(next_header)) = .false.
  endif
  !
  ! Set new occupant
  header_xnum(next_header) = kx
  ! Save back pointer
  v_header(kx) = next_header
  ! Save the header and user section
  call copy_header(head,header_content(next_header))
  call copy_user(user,user_content(next_header),error)
  call copy_assoc(assoc,assoc_content(next_header),error)
  if (error)  return
  ! All right, turn on flag
  got_header(kx) = .true.
  !
  ! Set next available slot
  if (next_header.eq.number_of_headers) then
    next_header = 1
  else
    next_header = next_header+1
  endif
  !
contains
  subroutine allocate_memory_header(error)
    use gkernel_interfaces
    use class_virtual
    logical, intent(inout) :: error
    ! Local
    type(header), target :: head
    real(kind=4) :: space_class
    integer(kind=8) :: memory_length
    integer(kind=4) :: ier
    !
    ! Use the logical name SPACE_CLASS which is the maximum number of MiB
    ! dedicated to the header+data buffers. Should be tuned to a fraction of the
    ! available RAM...
    space_class = 1.0
    ier = sic_ramlog('SPACE_CLASS',space_class)
    space_class = max(space_class/4,1.0)  ! 3/4 for data, 1/4 for headers. Minimum 1 MiB
    memory_length = space_class*1024*256 ! Memory length in 32 bits words
    ! The length of an header is fixed => Initialize it only once
    ! SIZEOF is not an Intrinsic => Use a work-around
    header_length = (locwrd(head%ends) - locwrd(head%start) + 8 ) / 4
    number_of_headers = memory_length/header_length
    ! Allocate memory
    allocate(header_content(number_of_headers),  &
             header_xnum(number_of_headers),     &
             user_content(number_of_headers),    &
             assoc_content(number_of_headers),   &
             stat=ier)
    if (failed_allocate(rname,'Header housekeeping arrays',ier,error))  return
    ! Initialize
    header_xnum(:) = 0
    next_header = 1
  end subroutine allocate_memory_header
end subroutine wheader_virtual
!
subroutine rheader_virtual(kx,head,user,assoc,error)
  use gbl_message
  use classic_api
  use class_interfaces, except_this=>rheader_virtual
  use class_types
  use class_virtual
  !---------------------------------------------------------------------
  ! @ private
  !  Read the input header from the virtual buffer
  !---------------------------------------------------------------------
  integer(kind=entry_length), intent(in)    :: kx     ! Entry number
  type(header),               intent(inout) :: head   ! The header
  type(class_user_t),         intent(inout) :: user   ! The user section
  type(class_assoc_t),        intent(inout) :: assoc  ! The Associated Arrays section
  logical,                    intent(inout) :: error  ! Error flag
  ! Local
  character(len=*), parameter :: rname='RHEADER_VIRTUAL'
  !
  if (.not.got_header(kx)) then
    call class_message(seve%e,rname,'Entry is not in virtual buffer')
    error = .true.
    return
  endif
  !
  call copy_header(header_content(v_header(kx)),head)
  call copy_user(user_content(v_header(kx)),user,error)
  call copy_assoc(assoc_content(v_header(kx)),assoc,error)
  if (error)  return
  !
end subroutine rheader_virtual
!
subroutine wdata_virtual(kx,ndata,data,error)
  use gildas_def
  use gbl_message
  use class_dependencies_interfaces
  use class_interfaces, except_this=>wdata_virtual
  use class_setup
  use class_virtual
  !---------------------------------------------------------------------
  ! @ private
  ! Copy the input data array to the virtual buffer
  !---------------------------------------------------------------------
  integer(kind=entry_length), intent(in)    :: kx       ! Entry number
  integer(kind=4),            intent(in)    :: ndata    ! Number of values to copy
  real(kind=4),               intent(in)    :: data(*)  ! Data array to be copied
  logical,                    intent(inout) :: error    ! Logical error flag
  ! Global
  include 'gbl_memory.inc'
  ! Local
  character(len=*), parameter :: rname='WDATA_VIRTUAL'
  character(len=message_length) :: mess
  integer(kind=8) :: current_length
  integer(kind=4) :: num_region,end_region,j,k
  integer(kind=address_length) :: ipv
  !
  if (ndata.le.0) then
     write(mess,'(A,I0,A)')  &
       'Internal error: input length is not positive (',ndata,')'
     call class_message(seve%e,rname,mess)
     error=.true.
     return
  endif
  !
  ! Allocate memory for CLASS data buffer at first call
  if (data_memory_length.eq.0) then
    call allocate_memory_data(error)
    if (error) then
      ! Could not allocate memory for data contents. Turn off the
      ! virtual buffer and leave
      error = .false.  ! Not an error
      call class_message(seve%w,rname,  &
        'Memory allocation failed, virtual buffer turned OFF')
      call setup_virtual_off(error)
      return
    endif
  endif
  !
  ! Major failure (see comment below...)
  current_length = max(ndata,smallest_length)
  if (current_length.gt.data_memory_length) then
    call class_message(seve%f,rname,'Buffer length exceeded, increase SPACE_CLASS')
    error = .true.
    return
  endif
  !
  ! This should treat the case where a single observation
  ! spans more than 1 region...
  !
  ! Check if we will overflow current region
  if (current_length+data_memory_offset .gt. region_length) then
     if (current_region.eq.n_regions) then
        current_region = 1
     else
        current_region = current_region + 1
     endif
     ! Might need several regions. Compute number of regions needed:
     num_region = (current_length + region_length - 1) / region_length
     end_region = current_region + num_region -1
     ! If it overflows at the end, put it in the first region:
     if (end_region.gt.n_regions) then
         current_region = 1
         end_region = current_region + num_region - 1
     endif
     !
     ! Free region range
     do k=current_region,end_region
        do j=1,number_of_obs(k)
          got_data(data_xnum(j,k)) = .false.
        enddo
        number_of_obs(k) = 0
     enddo
     data_memory_offset = 0  ! Restart at the beginning of the new selected region
  else
     end_region = current_region
  endif
  !
  ! Define return address
  v_data(kx) = data_memory_address + 4*(current_region-1)*region_length +  &
               4*data_memory_offset
  ! Set region occupancy
  do j = current_region,end_region
    number_of_obs(j) = number_of_obs(j)+1
    data_xnum(number_of_obs(j),j) = kx
  enddo
  ! Copy data
  ipv = gag_pointer(v_data(kx),memory)
  call r4tor4(data,memory(ipv),ndata)
  ! Set data occupancy
  got_data(kx) = .true.
  !
  ! Prepare next call.
  ! 1/ Set new current region
  current_region = end_region
  ! 2/ Increment memory offset. Since we may have written on several
  ! regions, it must be 'modulo'd.
  data_memory_offset = mod(data_memory_offset+current_length,region_length)
  ! 3/ If end_region (= current_region now) is full, we have to leave
  ! data_memory_offset to region_length. if-block above will detect this
  ! and search for a new region on next call.
  if (data_memory_offset.eq.0) then
     data_memory_offset=region_length
     if (set%virtual_check) virtual_bug_20081106=.true.
  endif
  !
contains
  subroutine allocate_memory_data(error)
    use class_dependencies_interfaces
    use class_virtual
    logical, intent(inout) :: error
    ! Local
    integer(kind=4) :: ier
    real(kind=4) :: space_class
    !
    ! Use the logical name SPACE_CLASS which is the maximum number of MiB
    ! dedicated to the header+data buffers. Should be tuned to a fraction of the
    ! available RAM...
    space_class = 1.0
    ier = sic_ramlog('SPACE_CLASS',space_class)
    space_class = max(3*space_class/4,1.0)  ! 3/4 for data, 1/4 for headers. Minimum 1 MiB
    data_memory_length = space_class*1024*256 ! Memory length in 32 bits words
    ! Get virtual memory
    error = (sic_getvm(data_memory_length,data_memory_address).ne.1)
    if (error) then
      data_memory_length = 0
      return
    endif
    !
    ! 'data_memory_length' is divided here in 'n_regions'. This optimization
    ! ensures that, when the buffer is full, we do not erase it completely
    ! when restarting from zero.
    !
    ! Initialize needed parameters.
    region_length = data_memory_length/n_regions  ! Size of a single region
    region_size = region_length/smallest_length
    data_memory_offset = 0                        ! Offset in current region
    current_region = 1                            ! First region used
    !
    allocate(number_of_obs(n_regions),data_xnum(region_size,n_regions),stat=ier)
    if (failed_allocate(rname,'Data housekeeping arrays',ier,error))  return
    number_of_obs(:) = 0  ! Number of observations stored in Nth region.
  end subroutine allocate_memory_data
  !
end subroutine wdata_virtual
!
subroutine rdata_virtual(kx,data,ndata,error)
  use gbl_message
  use classic_api
  use class_dependencies_interfaces
  use class_interfaces, except_this=>rdata_virtual
  use class_virtual
  !---------------------------------------------------------------------
  ! @ private
  !---------------------------------------------------------------------
  integer(kind=entry_length), intent(in)    :: kx       ! Entry number
  integer(kind=4),            intent(in)    :: ndata    ! Number of values to read
  real(kind=4),               intent(out)   :: data(*)  ! Data array to be filled
  logical,                    intent(inout) :: error    ! Logical error flag
  ! Global
  include 'gbl_memory.inc'
  ! Local
  character(len=*), parameter :: rname='RDATA'
  integer(kind=address_length) :: ipv
  !
  if (virtual_bug_20081106) then
    call class_message(seve%e,rname,'You were subject to the virtual buffer bug')
    call class_message(seve%e,rname,'See HELP SET VIRTUAL for details')
    virtual_bug_20081106 = .false.
    error = .true.
    return
  endif
  !
  ! Transfer data from memory to obs
  ipv = gag_pointer(v_data(kx),memory)
  call r4tor4(memory(ipv),data,ndata)
  !
end subroutine rdata_virtual
!
subroutine virtual_reset_headers
  use class_index
  use class_virtual
  !---------------------------------------------------------------------
  ! @ private
  ! Unmap all headers
  !---------------------------------------------------------------------
  got_header(1:ix%mobs) = .false.
  !
end subroutine virtual_reset_headers
!
subroutine virtual_reset_data
  use class_index
  use class_virtual
  !---------------------------------------------------------------------
  ! @ private
  ! Unmap all data
  !---------------------------------------------------------------------
  got_data(1:ix%mobs) = .false.
  !
end subroutine virtual_reset_data
!
subroutine virtual_new_file(error)
  use class_interfaces, except_this=>virtual_new_file
  use class_index
  use class_virtual
  !---------------------------------------------------------------------
  ! @ private
  ! A New File : deassign IO channel, unmap all sections, open again
  !---------------------------------------------------------------------
  logical, intent(out) :: error
  !
  error = .false.
  call reallocate_virtual(ix%mobs,.false.,error)
  if (error)  return
  call virtual_reset_headers
  call virtual_reset_data
  !
  virtual_bug_20081106 = .false.
  data_memory_offset = 0
  !
  if (allocated(number_of_obs)) number_of_obs(:) = 0
  if (allocated(header_xnum))   header_xnum(:) = 0
end subroutine virtual_new_file
!
subroutine reallocate_virtual(mobs,keep,error)
  use gildas_def
  use gkernel_interfaces
  use class_virtual
  !---------------------------------------------------------------------
  ! @ private
  !---------------------------------------------------------------------
  integer(kind=entry_length), intent(in)    :: mobs   !
  logical,                    intent(in)    :: keep   ! Keep previous values?
  logical,                    intent(inout) :: error  !
  ! Local
  character(len=*), parameter :: rname='ALLOCATE'
  integer(kind=entry_length) :: old_mobs
  logical :: mykeep
  integer(kind=4) :: ier
  logical, allocatable :: bufl4(:)
  integer(kind=address_length), allocatable :: bufad(:)
  !
  if (allocated(got_header)) then
    old_mobs = size(got_header,kind=entry_length)
    if (old_mobs.ge.mobs)  return
    mykeep = keep
  else
    old_mobs = 0
    mykeep = .false.  ! No previous allocation => nothing to keep
  endif
  !
  if (mykeep) then
    allocate(bufl4(old_mobs),bufad(old_mobs),stat=ier)
    if (failed_allocate(rname,'tmp workspaces',ier,error))  return
  endif
  !
  call reallocate_virtual_l4('got_header',got_header,error)
  if (error)  return
  call reallocate_virtual_l4('got_data',got_data,error)
  if (error)  return
  call reallocate_virtual_ad('v_header',v_header,error)
  if (error)  return
  call reallocate_virtual_ad('v_data',v_data,error)
  if (error)  return
  !
  if (mykeep)  deallocate(bufl4,bufad)
  !
contains
  subroutine reallocate_virtual_l4(name,val,error)
    character(len=*), intent(in)    :: name    ! Name of array
    logical,          allocatable   :: val(:)  ! The array
    logical,          intent(inout) :: error   !
    ! Local
    integer(kind=4) :: ier
    !
    if (mykeep)  bufl4(:) = val(1:old_mobs)
    if (allocated(val)) deallocate(val)
    allocate(val(mobs),stat=ier)
    if (failed_allocate(rname,name,ier,error))  return
    if (mykeep) then
      val(1:old_mobs) = bufl4(:)
      val(old_mobs:mobs) = .false.
    else
      val(1:mobs) = .false.
    endif
  end subroutine reallocate_virtual_l4
  subroutine reallocate_virtual_ad(name,val,error)
    use gildas_def
    character(len=*),             intent(in)    :: name    ! Name of array
    integer(kind=address_length), allocatable   :: val(:)  ! The array
    logical,                      intent(inout) :: error   !
    ! Local
    integer(kind=4) :: ier
    !
    if (mykeep)  bufad(:) = val(1:old_mobs)
    if (allocated(val)) deallocate(val)
    allocate(val(mobs),stat=ier)
    if (failed_allocate(rname,name,ier,error))  return
    if (mykeep) then
      val(1:old_mobs) = bufad(:)
      val(old_mobs:mobs) = 0
    else
      val(1:mobs) = 0
    endif
  end subroutine reallocate_virtual_ad
end subroutine reallocate_virtual
!
subroutine deallocate_virtual(error)
  use gbl_message
  use class_dependencies_interfaces
  use class_interfaces, except_this=>deallocate_virtual
  use class_virtual
  !---------------------------------------------------------------------
  ! @ private
  !---------------------------------------------------------------------
  logical, intent(inout) :: error  !
  ! Local
  character(len=*), parameter :: rname='DEALLOCATE'
  integer(kind=4) :: ier
  !
  call class_message(seve%d,rname,'Deallocate virtual memory')
  !
  if (allocated(got_header)) then
    deallocate(got_header,stat=ier)
    if (ier.ne.0) then
      call class_message(seve%f,rname,'Could not deallocate got_header')
      error = .true.
    endif
  endif
  !
  if (allocated(got_data)) then
    deallocate(got_data,stat=ier)
    if (ier.ne.0) then
      call class_message(seve%f,rname,'Could not deallocate got_data')
      error = .true.
    endif
  endif
  !
  if (allocated(v_header)) then
    deallocate(v_header,stat=ier)
    if (ier.ne.0) then
      call class_message(seve%f,rname,'Could not deallocate v_header')
      error = .true.
    endif
  endif
  !
  if (allocated(v_data)) then
    deallocate(v_data,stat=ier)
    if (ier.ne.0) then
      call class_message(seve%f,rname,'Could not deallocate v_data')
      error = .true.
    endif
  endif
  !
  call deallocate_virtual_contents(error)
  !
end subroutine deallocate_virtual
!
subroutine deallocate_virtual_contents(error)
  use gbl_message
  use class_dependencies_interfaces
  use class_interfaces, except_this=>deallocate_virtual_contents
  use class_virtual
  !---------------------------------------------------------------------
  ! @ private
  !  Deallocate the virtual buffers associated to contents (headers and
  ! data), i.e. those allocated on SPACE_CLASS
  !---------------------------------------------------------------------
  logical, intent(inout) :: error
  ! Local
  character(len=*), parameter :: rname='VIRTUAL'
  integer(kind=4) :: ier,iassoc
  !
  if (allocated(header_content)) then
    ! Clean user sections
    ! ZZZ memory leak here: we should first clean all the user sections
    ! with a deallocate_user!
    ! Clean associated arrays section
    do iassoc=1,size(assoc_content)
      call deallocate_assoc(assoc_content(iassoc),error)
    enddo
    deallocate(header_content,header_xnum,user_content,assoc_content,stat=ier)
    if (ier.ne.0) then
      call class_message(seve%e,rname,'Could not deallocate header_memory')
      error = .true.
    endif
  endif
  !
  if (data_memory_length.ne.0) then
    call free_vm(data_memory_length,data_memory_address)
    deallocate(number_of_obs,data_xnum,stat=ier)
    data_memory_length = 0
    if (ier.ne.0) then
      call class_message(seve%e,rname,'Could not deallocate data_memory')
      error = .true.
    endif
  endif
  !
end subroutine deallocate_virtual_contents
!
subroutine setup_virtual_off(error)
  use class_interfaces, except_this=>setup_virtual_off
  use class_setup
  !---------------------------------------------------------------------
  ! @ private
  !  Turn off the SET VIRTUAL buffer and free the associated
  ! SPACE_CLASS memory
  !---------------------------------------------------------------------
  logical, intent(inout) :: error
  !
  if (set%virtual) then
    call virtual_reset_headers             ! Forget saved headers
    call virtual_reset_data                ! Forget saved data
  endif
  call deallocate_virtual_contents(error)  ! Free SPACE_CLASS
  set%virtual = .false.
  !
end subroutine setup_virtual_off
