!
subroutine uv_sort (error,sorted,shift,new,uvmax,uvmin)
  use gkernel_interfaces
  use mapping_interfaces, except_this=>uv_sort
  use clean_def
  use clean_arrays
  use gbl_message
  !----------------------------------------------------------------------
  ! @ private
  !
  ! MAPPING
  !   Sort the input UV table
  !----------------------------------------------------------------------
  logical, intent(inout) :: sorted      ! Is table sorted ?
  logical, intent(inout) :: shift       ! Do we shift phase center ?
  logical, intent(out) :: error
  real(kind=8), intent(inout) :: new(3) ! New phase center and PA
  real, intent(out) :: uvmin            ! Min baseline
  real, intent(out) :: uvmax            ! Max baseline
  !
  ! Global variables:
  real(8), parameter :: pi=3.14159265358979323846d0
  real(8), parameter :: f_to_k = 2.d0*pi/299792458.d-6
  real(kind=8) freq, off(3)
  real pos(2), cs(2)
  integer nu,nv,ier
  real, pointer :: duv_previous(:,:)
  real, pointer :: duv_next(:,:)
  !
  ! The UV table is available in HUV%
  if (huv%loca%size.eq.0) then
    call map_message(seve%e,'UV_MAP','No UV data loaded')
    error = .true.
    return
  endif
  nu = huv%gil%dim(1)
  nv = huv%gil%dim(2)
  !
  ! Correct for new phase center if required
  if (shift) then
    if (huv%gil%ptyp.eq.p_none) then
      call map_message(seve%w,'SHIFT','No previous phase center info')
      huv%gil%a0 = huv%gil%ra
      huv%gil%d0 = huv%gil%dec
      huv%gil%pang = 0.d0
      huv%gil%ptyp = p_azimuthal
    elseif (huv%gil%ptyp.ne.p_azimuthal) then
      call map_message(seve%w,'SHIFT','Previous projection type not SIN')
      huv%gil%ptyp = p_azimuthal
    endif
    call uv_shift (new,huv%gil%a0,huv%gil%d0,huv%gil%pang,   &
        &      off,shift)
    huv%gil%posi_words = def_posi_words
    huv%gil%proj_words = def_proj_words
  endif
  if (shift) then
    sorted = .false.
  else
    call check_order (duv,nu,nv,sorted)
  endif
  !
  ! Get center frequency
  freq = gdf_uv_frequency(huv,huv%gil%ref(1))
  !
  ! If already sorted, use it
  if (sorted) then
    call map_message(seve%i,'UVSORT','UV table is already sorted')
    !
    ! Compute UVMAX
    call uvgmax (huv,duv,uvmax,uvmin)
    !
    ! Else, create another copy
  else
    call map_message(seve%i,'UVSORT','Sorting UV table...')
    !
    ! Compute observing frequency, and new phase center in wavelengths
    if (shift) then
      huv%gil%a0 = new(1)
      huv%gil%d0 = new(2)
      huv%gil%pang = new(3)
      cs(1)  =  cos(off(3))
      cs(2)  = -sin(off(3))
      ! Note that the new phase center is counter-rotated because rotations
      ! are applied before phase shift.
      pos(1) = - freq * f_to_k * ( off(1)*cs(1) - off(2)*cs(2) )
      pos(2) = - freq * f_to_k * ( off(2)*cs(1) + off(1)*cs(2) )
    else
      pos(1) = 0.0
      pos(2) = 0.0
      cs(1) = 1.0
      cs(2) = 0.0
    endif
    !
    !$ call uv_dump_buffers('UV_SORT')
    !
    ! OK, rotate, shift, sort and copy...
    nullify (duv_previous, duv_next)
    call uv_find_buffers ('UV_SORT',nu,nv, duv_previous, duv_next, error)
    if (error) return
    call uvsort_uv (nu,nv,huv%gil%ntrail,duv_previous,duv_next,   &
         &        pos,cs,uvmax,uvmin,error)
    call uv_clean_buffers (duv_previous, duv_next,error)
    if (error) return
  endif
  !
  ! Now transform UVMAX in kiloWavelength (including 2 pi factor)
  uvmax = uvmax*freq*f_to_k
  uvmin = uvmin*freq*f_to_k
  error = .false.
end subroutine uv_sort
!
!
subroutine one_beam (rname,map,huv,hbeam,hdirty,   &
     &    nx,ny,nu,nv,uvdata,   &
     &    w_mapu, w_mapv, w_grid, w_weight, w_v, do_weig,    &
     &    wcol,mcol,wfft,sblock,cpu0,error,uvmax)
  use gkernel_interfaces
  use mapping_interfaces, except_this=>one_beam
  use clean_def
  use image_def
  use gbl_message
  use clean_default
  !------------------------------------------------------------------------
  ! @ private
  !
  ! MAPPING
  !   Compute a map from a CLIC UV Sorted Table
  !   by Gridding and Fast Fourier Transform, with
  !   one single beam for all channels.
  !
  ! Input :
  ! a precessed UV table, sorted in V, ordered in
  ! (U,V,W,D,T,iant,jant,nchan(real,imag,weig))
  ! Output :
  ! a beam image
  ! a VLM cube
  ! Work space :
  ! a  VLM complex Fourier cube (first V value is for beam)
  !------------------------------------------------------------------------
  character(len=*), intent(in) :: rname   ! Calling Task name
  type (uvmap_par), intent(inout) :: map  ! Mapping parameters
  type (gildas), intent(inout) :: huv     ! UV data set
  type (gildas), intent(inout) :: hbeam   ! Dirty beam data set
  type (gildas), intent(inout) :: hdirty  ! Dirty image data set
  integer, intent(in) :: nx   ! X size
  integer, intent(in) :: ny   ! Y size
  integer, intent(in) :: nu   ! Size of a visibilities
  integer, intent(in) :: nv   ! Number of visibilities
  real, intent(inout) :: uvdata(nu,nv)
  real, intent(inout) :: w_mapu(nx)      ! U grid coordinates
  real, intent(inout) :: w_mapv(ny)      ! V grid coordinates
  real, intent(inout) :: w_grid(nx,ny)   ! Gridding space
  real, intent(inout) :: w_weight(nv)    ! Weight of visibilities
  real, intent(inout) :: w_v(nv)         ! V values
  logical, intent(inout) :: do_weig
  !
  real, intent(inout) :: wfft(*)     ! Work space
  real, intent(inout) :: cpu0        ! CPU
  real, intent(inout) :: uvmax       ! Maximum baseline
  integer, intent(inout) :: sblock   ! Blocking factor
  integer, intent(inout) :: wcol     ! Weight channel
  integer, intent(inout) :: mcol(2)  ! First and last channel
  logical, intent(inout) :: error
  !
  if (omp_single_beam.eq.0) then
    Print *,'Calling SERIAL one_beam'
    call one_beam_serial (rname,map,huv,hbeam,hdirty,   &
     &    nx,ny,nu,nv,uvdata,   &
     &    w_mapu, w_mapv, w_grid, w_weight, w_v, do_weig,    &
     &    wcol,mcol,wfft,sblock,cpu0,error,uvmax)
  else
    Print *,'Calling PARALLEL one_beam'
    call one_beam_para (rname,map,huv,hbeam,hdirty,   &
     &    nx,ny,nu,nv,uvdata,   &
     &    w_mapu, w_mapv, w_grid, w_weight, w_v, do_weig,    &
     &    wcol,mcol,wfft,sblock,cpu0,error,uvmax)
  endif
end subroutine one_beam
!
subroutine check_order(visi,np,nv,sorted)
  !----------------------------------------------------------
  ! @ private
  !
  ! MAPPING
  !   Check if visibilites are sorted.
  !   Chksuv does a similar job, but using V values and an index.
  !----------------------------------------------------------
  integer, intent(in) :: np       ! Size of a visibility
  integer, intent(in) :: nv       ! Number of visibilities
  real, intent(in) :: visi(np,nv) ! Visibilities
  logical, intent(out) :: sorted
  !
  real vmax
  integer iv
  !
  vmax = visi(2,1)
  do iv=1,nv
    if (visi(2,iv).lt.vmax) then
      sorted = .false.
      return
    endif
    vmax = visi(2,iv)
  enddo
  sorted = .true.
end subroutine check_order
!
subroutine map_parameters(task,map,freq,uvmax,uvmin,uvma,map_range)
  use gkernel_interfaces
  use mapping_interfaces, only : map_message
  use clean_def
  use clean_arrays
  use clean_types
  use clean_default
  use gbl_message
  !----------------------------------------------------------------------
  ! @ private-mandatory
  !
  ! Prepare the MAP parameters for UV_MAP, MX or UV_RESTORE
  ! Reads this from SIC variables
  !-----------------------------------------------------------------------
  character(len=*), intent(in)    :: task ! Input task name (UV_MAP or MX)
  type (uvmap_par), intent(inout) :: map  ! Map parameters
  real(8), intent(inout) :: freq          ! Observing frequency
  real(4), intent(in) :: uvmax, uvmin     ! Min & Max UV in m
  real(4), intent(out) :: uvma            ! Max UV in kLambda
  real(4), intent(in), optional :: map_range(2)  ! Range of offsets
  !
  real(8), parameter :: pi=3.14159265358979323846d0
  real(8), parameter :: f_to_k = 2.d0*pi/299792458.d-6
  real, parameter :: pixel_per_beam=5.0
  character(len=80) :: chain
  real uvmi
  integer :: i, ms, k
  type (uvmap_par), save :: last_map
  !
  freq = gdf_uv_frequency(huv,huv%gil%ref(1))
  uvma = uvmax/(freq*f_to_k)
  uvmi = uvmin/(freq*f_to_k)
  write(chain,100) huv%gil%dim(2),huv%gil%nchan
  call map_message(seve%i,task,chain)
  write(chain,101) uvmi,uvma,' meters'
  call map_message(seve%i,task,chain)
  write(chain,101) uvmin*1e-3/2.0/pi,uvmax*1e-3/2.0/pi,' kiloWavelength'
  call map_message(seve%i,task,chain)
  !
  ! Now get MAP_CELL and MAP_SIZE if Zero.
  if (map%xycell(1).eq.0) then
    map%xycell(1) = 0.01*nint(360.0*3600.0*100.0/uvmax/pixel_per_beam)
    if (map%xycell(1).le.0.02) then
      map%xycell(1) =   &
           &        0.002*nint(360.0*3600.0*500.0/uvmax/pixel_per_beam)
    endif
    map%xycell(2) = map%xycell(1)  ! In ", rounded to 0.02 or 0.002"
  endif
  !
  ! Get a wide enough field of view...
  if (map%field(1).ne.0) then
    if (map%field(2).eq.0) map%field(2) = map%field(1)
    map%size = map%field/map%xycell !
    map%size =   2**nint(log(real(map%size))/log(2.0))
    !
    map%xycell = sqrt(map%field(1)*map%field(2)/(map%size(1)*map%size(2)))
    !
  endif
  !
  !
  if (map%size(1).eq.0) then
    ! This is an obsolete definition of the Primary beam
    write(chain,'(A,F11.1,A)') 'Primary beam size ', &
        &  huv%gil%majo*180*3600/pi,'"'
    call map_message(seve%i,task,chain)
    !
    map%size = huv%gil%majo/(map%xycell*pi/180/3600)
    if (present(map_range)) map%size = map%size + map_range/(map%xycell*pi/180/3600)
    !!Print *,map%size(2), nint(pixel_per_beam*uvmax/uvmin)
    map%size(1) = max(map%size(1),nint(pixel_per_beam*uvmax/uvmin))
    map%size(2) = max(map%size(2),nint(pixel_per_beam*uvmax/uvmin))
    !
    ! One could do that with Log_2
    do k=1,2
      ms = 32
      do
        if (1.20*ms.lt.map%size(k)) then
          ms = 2*ms
        else
          exit
        endif
      enddo
      map%size(k) = ms
    enddo
    write(chain,'(A,F11.1,A)') 'Image size ',map%size(1)*map%xycell(1),'" '
    call map_message(seve%i,task,chain)
  endif
  !
  write(chain,103) map%size
  call map_message(seve%i,task,chain)
  write(chain,104) map%xycell
  call map_message(seve%i,task,chain)
  !
  ! Check whether Weights have to be recomputed
  if (map%mode.ne.last_map%mode) do_weig = .true.
  if (map%mode.ne.'NATURAL ') then
    if (map%uniform(1).ne.last_map%uniform(1)) do_weig = .true.
    if (map%uniform(2).ne.last_map%uniform(2)) do_weig = .true.
  endif
  if (map%ctype.ne.last_map%ctype) do_weig = .true.
  do i=1,4
    if (map%taper(i).ne.last_map%taper(i)) do_weig = .true.
  enddo
  last_map = map
  !
  ! Subroutine DOWEI uses map%uniform = 0.0 for Natural Weighting
  if (map%mode.eq.'NATURAL ') map%uniform(1) = 0
  !
100 format('Found ',i12,' Visibilities, ',i4,' channels')
101 format('Baselines ',f9.1,' - ',f9.1,a)
103 format('Map size is   ',i4,' by ',i4)
104 format('Pixel size is ',f8.3,' by ',f8.3,'"')
end subroutine map_parameters
!
subroutine cct_prepare(nc,a_method,task,error)
  use gkernel_interfaces
  use clean_def
  use clean_arrays
  use clean_types
  use clean_default
  use gbl_message
  !---------------------------------------------------
  ! @ private
  !
  ! MAPPING
  !   Prepare the Clean Component Table
  !---------------------------------------------------
  integer, intent(in) :: nc      ! Number of channels
  type(clean_par), intent(in) :: a_method   ! Input method
  character(len=*), intent(in) :: task      ! Caller's name
  logical, intent(out) :: error  !
  !
  integer c_iter,ier,x_iter
  real, allocatable :: dtmp(:,:,:)
  character(len=12) :: number
  !
  error = .false.
  x_iter = a_method%m_iter
  if (a_method%method.eq.'MULTI') x_iter = a_method%ninflate*x_iter
  !print *,'CCT-Prepare, Method ',method%method,m_iter,x_iter
  !print *,'CCT-Prepare, FIRST ',a_method%first,' LAST ',a_method%last
  if (allocated(dcct)) then
    c_iter = ubound(dcct,3)
    if (c_iter.lt.x_iter) then
      write(number,'(i12)') x_iter
      call map_message(seve%i,task,'Increasing size of current CCT table to ' &
      & //adjustl(number))
      allocate(dtmp(3,nc,c_iter),stat=ier)
      if (ier.eq.0) then
        dtmp(:,:,:) = dcct
        deallocate(dcct)
        allocate(dcct(3,nc,x_iter),stat=ier)
        if (ier.ne.0) then
          call map_message(seve%e,task,'Memory allocation failure')
          error = .true.
          return
        endif
        dcct(:,:,1:c_iter) = dtmp
        dcct(:,:,c_iter+1:) = 0
        deallocate(dtmp)
        c_iter = x_iter
        hcct%loca%size = 3*nc*x_iter
      else
        call map_message(seve%e,task,'Memory allocation failure')
        error = .true.
        return
      endif
    else if (c_iter.ge.x_iter) then
      call map_message(seve%i,task,'Re-using current CCT table')
      dcct(:,a_method%first:a_method%last,:) = 0.0
    endif
  else
    write(number,'(i12)') x_iter
    call map_message(seve%i,task,'Allocating new CCT table of size '// &
      & adjustl(number))
    allocate(dcct(3,nc,x_iter),stat=ier)
    if (ier.ne.0) then
      call map_message(seve%e,task,'Memory allocation failure')
      error = .true.
      return
    endif
    dcct = 0.0
    c_iter = x_iter
    hcct%loca%size=3*nc*x_iter
  endif
  !
  ! Define the image header
  call gildas_null(hcct)
  call gdf_copy_header (hdirty,hcct,error)
  hcct%gil%ndim = 3
  hcct%char%unit = 'Jy'
  !
  hcct%gil%dim(1) = 3
  ! Keep the same axis description
  hcct%gil%xaxi = 1
  !
  hcct%gil%convert(:,2) = hdirty%gil%convert(:,3)
  hcct%gil%convert(:,3) = hdirty%gil%convert(:,2)
  hcct%gil%dim(2) = hdirty%gil%dim(3)
  hcct%char%code(2) = hdirty%char%code(3)
  hcct%gil%faxi = 2
  hcct%gil%dim(3) = c_iter
  hcct%char%code(3) = 'COMPONENT'
  hcct%gil%yaxi = 3
  !
  ! Initialize BLCs...
  hcct%blc = 0
  hcct%trc = 0
  ! Cleanup
end subroutine cct_prepare
!

