subroutine map_copy_par(in,out)
  use clean_def
  !-----------------------------------------------------------------------
  !
  ! @ private
  !   Copy a MAP structure to another one, but avoid erasing
  !   the number of fields in it. The result must have an intent(inout)
  !   to avoid erasing the allocatable array in the derived type.
  !
  !-----------------------------------------------------------------------
  type(uvmap_par), intent(in) :: in
  type(uvmap_par), intent(inout) :: out
  !
  out%taper = in%taper         ! UV Taper
  out%size = in%size           ! Map size
  out%xycell = in%xycell       ! Map cell
  out%uvcell = in%uvcell       ! UV cell
  out%uniform = in%uniform     ! Weighting parameters
  out%wcol = in%wcol           ! Weighting channel
  out%ctype = in%ctype         ! Convolution mode
  out%support = in%support     ! Support of gridding function
  out%beam = in%beam           ! One beam every N channels
  out%field = in%field         ! Field of view in arcsecond
  out%precis = in%precis       ! Precision at map edges
  out%mode = in%mode           ! Weighting mode
  out%shift = in%shift         ! Shift or rotate UV data
  out%ra_c = in%ra_c           ! Right ascension
  out%dec_c = in%dec_c         ! Declination
  out%angle = in%angle         ! New position angle of UV data
  out%truncate = in%truncate   ! Truncation radius
  !
end subroutine map_copy_par
!
subroutine map_prepare(task,map,error)
  use clean_def
  use clean_default
  use gkernel_interfaces
  use gbl_message
  !
  ! @ private
  !
  character(len=*), intent(in) :: task
  type(uvmap_par), intent(inout) :: map
  logical, intent(out) :: error
  ! character(len=*), parameter :: obsol='obsolete'   ! Later
  character(len=*), parameter :: obsol='obsolescent'
  !
  ! Check version
  if (map_version.lt.-1 .or. map_version.gt.1) then
    call map_message(seve%e,task,'Invalid MAP_VERSION, should be -1,0 or 1')
    error = .true.
    return
  endif
  !
  ! Check old syntax
  if (old_map%uniform(1).ne.save_map%uniform(1)) then
    call map_message(seve%w,task,'UV_CELL is '//obsol//', use MAP_UVCELL instead')
    default_map%uniform(1) = old_map%uniform(1)
  endif
  if (old_map%uniform(2).ne.save_map%uniform(2)) then
    call map_message(seve%w,task,'UV_CELL is '//obsol//', use MAP_ROBUST instead')
    default_map%uniform(2) = old_map%uniform(2)
  endif
  if (old_map%taper(4).ne.save_map%taper(4)) then
    call map_message(seve%w,task,'TAPER_EXPO is '//obsol//', use MAP_TAPER_EXPO instead')
    save_map%taper(4) = old_map%taper(4)
    default_map%taper(4) = old_map%taper(4)
  endif
  if (any(old_map%taper.ne.save_map%taper)) then
    call map_message(seve%w,task,'UV_TAPER is '//obsol//', use MAP_UVTAPER instead')
    default_map%taper = old_map%taper
  endif
  if (old_map%ctype.ne.save_map%ctype) then
    call map_message(seve%w,task,'CONVOLUTION is '//obsol//', use MAP_CONVOLUTION instead')
    default_map%ctype = old_map%ctype
  endif
  if (old_map%mode.ne.save_map%mode) then
    call map_message(seve%w,task,'WEIGHT_MODE is '//obsol//', use MAP_WEIGHT instead')
    default_map%mode = old_map%mode
  endif
  if (old_map%shift.neqv.save_map%shift) then
    call map_message(seve%w,task,'UV_SHIFT is '//obsol//', use MAP_SHIFT instead')
    default_map%shift = old_map%shift
  endif
  !
  ! Copy the current default to the actual structure
  call map_copy_par(default_map,map)
  !
  ! Weighting scheme
  call get_weightmode(task,map%mode,error)
  if (error) return
  default_map%mode = map%mode ! Solve ambiguities
  !
  if (map%mode.eq.'NATU') map%uniform(1) = 0.0
  if (map%uniform(2).le.0.0) map%uniform(2) =  1.0
  !
  error = .false.
  !
  ! Here, we do not care about the fields issue
  old_map = default_map
  save_map = default_map
end subroutine map_prepare
!
subroutine sub_mosaic(name,error)
  use gkernel_interfaces
  use mapping_interfaces, except_this=>sub_mosaic
  use clean_def
  use clean_arrays
  use gbl_message
  !----------------------------------------------------------------------
  ! @ private
  !
  ! MAPPING     MOSAIC ON|OFF
  !             Activates or desactivates the mosaic mode
  !----------------------------------------------------------------------
  character(len=*), intent(in) :: name
  logical, intent(out) :: error
  !
  real(8), parameter :: pi=3.14159265358979323846d0
  !
  integer nf
  logical mosaic, old_mosaic
  real prim_beam
  character(len=message_length) :: mess
  character(len=6) :: rname = 'MOSAIC'
  !
  old_mosaic = user_method%mosaic
  mosaic = name.eq.'ON'
  !
  if (mosaic) then
    if (old_mosaic) then
      call map_message(seve%i,rname,'Already in MOSAIC mode')
    else
      call map_message(seve%i,rname,'Switch to MOSAIC mode')
      call gprompt_set('MOSAIC')
    endif
    if (user_method%trunca.ne.0.) then
      nf = hprim%gil%dim(1)
      ! Should that be replaced by something else ? e.g. in the Telescope Section ?
      prim_beam = hprim%gil%convert(3,4)   ! convention
      write(mess,100) 'Last mosaic loaded: ', nf,' fields'
      call map_message(seve%i,rname,mess)
      write(mess,101) 'Primary beam (arcsec) = ',prim_beam*180*3600/pi
      call map_message(seve%i,rname,mess)
      write(mess,101) 'Truncation level B_MIN = ',user_method%trunca
      call map_message(seve%i,rname,mess)
    else
      call map_message(seve%w,rname,'No mosaic loaded so far')
    endif
    write(mess,101) 'Current value: SEARCH_W = ',user_method%search
    call map_message(seve%i,rname,mess)
    write(mess,101) 'Current value: RESTORE_W = ',user_method%restor
    call map_message(seve%i,rname,mess)
    user_method%mosaic = .true.
  else
    if (.not.old_mosaic) then
      call map_message(seve%i,rname,'Already in NORMAL mode')
    else
      call map_message(seve%i,rname,'Switch to NORMAL mode')
      call gprompt_set('MAPPING')
      user_method%trunca = 0.0
    endif
    user_method%mosaic = .false.
  endif
  !
100 format(a,i3,a)
101 format(a,f5.2)
end subroutine sub_mosaic
!
subroutine get_bsize(h,rname,bsize,btrunc,line,o_teles,o_trunc,error)
  use image_def
  use gkernel_interfaces
  use gbl_message
  !---------------------------------------------------------------------
  ! @ private
  !
  ! MAPPING  Support for command
  !   PRIMARY [BeamSize] [/TELESCOPE ALMA|ACA|NOEMA] [/TRUNCATE Percent]
  !   UV_MAP  [BeamSize] [/TELESCOPE ALMA|ACA|NOEMA] [/TRUNCATE Percent]
  !
  ! Return the primary beam size in radian
  !---------------------------------------------------------------------
  type(gildas), intent(in) :: h          ! UV data header
  character(len=*), intent(in) :: rname  ! Caller's name
  real(4), intent(out) :: bsize          ! Beam size in radian
  real(4), intent(out) :: btrunc         ! Truncation level [0,1]
  character(len=*), intent(in) :: line   ! Command line
  integer, intent(in) :: o_teles         ! Telescope option number
  integer, intent(in) :: o_trunc         ! Truncation option number
  logical, intent(out) :: error          ! Error flag
  !
  real(8), parameter :: pi=3.14159265358979323846d0
  real(8), parameter :: clight=299792458d0
  integer, parameter :: mteles=6
  character(len=12) :: cteles,kteles,vteles(mteles)
  character(len=60) :: chain
  integer :: i, ier, nc, iteles
  real :: factors(mteles), diams(mteles), diamet, factor, beamsize
  data vteles /'ACA','ALMA','ATCA','NOEMA','SMA','JVLA'/
  data diams  /7.,12.,20.,15.,6.,25./
  data factors /1.13,1.13,1.13,1.22,1.13,1.13/
   !
  ! Verify if any Deconvolved data
  !
  if (h%loca%size.eq.0) then
    call map_message(seve%e,rname,'No input data')
    error = .true.
    return
  endif
  !
  if (h%gil%nteles.eq.1) then
    ! See if it is a recognized telescope
    cteles = h%gil%teles(1)%ctele
    call map_message(seve%i,rname,'Found telescope '//trim(cteles)//' from data')
    diamet = h%gil%teles(1)%diam
    call sic_ambigs(' ',cteles,kteles,iteles,vteles,mteles,error)
    if (error) then
      factor = 1.13
      call map_message(seve%w,rname,'Using default beam 1.13 Lambda/D')
      error = .false.
    else
      factor = factors(iteles)
    endif
    beamsize = factor*(clight/h%gil%freq/1d6)/diamet
  else
    beamsize = 0
  endif
  !
  if (o_teles.ne.0 .and. sic_present(o_teles,0)) then
    call sic_ke(line,o_teles,1,cteles,nc,.true.,error)
    if (error) return
    call sic_ambigs(rname,cteles,kteles,iteles,vteles,mteles,error)
    if (error) return
    bsize = factors(iteles)*(clight/h%gil%freq/1d6)/diams(iteles)
  else if (sic_present(0,1)) then
    call sic_r4(line,0,1,bsize,.true.,error)
    if (error) return
    bsize = bsize*pi/180/3600
  else if (h%gil%majo.ne.0) then
    if (error) return
    bsize = h%gil%majo
  else if (beamsize.eq.0) then
    call map_message(seve%e,rname,'No primary beam from data or as argument')
    error = .true.
    return
  endif
  !
  if (abs(beamsize-bsize).gt.0.02*bsize) then
    write(chain,'(A,F8.1,A)') 'Specified beam differs from value in data ',beamsize*180*3600/pi,'"'
    call map_message(seve%w,rname,chain)
  endif
  write(chain,'(A,F8.1,A)') 'Primary beam ',bsize*180*3600/pi,'"'
  call map_message(seve%i,rname,chain)
  !
  ! Truncation
  call sic_get_real('MAP_TRUNCATE',btrunc,error)
  call sic_r4(line,o_trunc,1,btrunc,.false.,error)
  btrunc = btrunc*0.01
  !
end subroutine get_bsize
!
subroutine primary_comm(line,error)
  use gkernel_interfaces
  use mapping_interfaces, except_this=>primary_comm
  use clean_def
  use clean_arrays
  use clean_types
  use clean_default
  use gbl_message
  !------------------------------------------------------------------------
  ! @ private
  !
  ! MAPPING  Support for command
  !   PRIMARY [BeamSize] [/TELESCOPE ALMA|ACA|NOEMA][/TRUNCATE Percent]
  !
  ! Correct for primary beam
  !------------------------------------------------------------------------
  character(len=*), intent(in) :: line ! Command line
  logical, intent(out) :: error
  !
  character(len=*), parameter :: rname='PRIMARY'
  real(8), parameter :: pi=3.14159265358979323846d0
  integer, parameter :: o_teles=1
  integer, parameter :: o_trunc=2
  !
  real, allocatable :: dflat(:,:)
  integer :: i, ier, nc
  real :: btrunc
  real :: bsize
  character(len=80) :: mess
  !
  ! Verify if any Deconvolved data and get Beam size
  !
  call get_bsize(hclean,rname,bsize,btrunc,line,o_teles,o_trunc,error)
  if (error) return
  !
  write(mess,'(a,f10.2,a,f6.0,a)') 'Correcting for a beam size of ',&
    & bsize/pi*180*3600,'" down to ',100*btrunc,'% '
  call map_message(seve%i,rname,mess)
  !
  if (allocated(dsky)) then
    if (any(hsky%gil%dim.ne.hclean%gil%dim)) then
      Print *,'Programming error - Sky & Clean mismatch'
      deallocate(dsky)
    else
      call map_message(seve%i,rname,'Re-using sky memory')
      call sic_delvariable('SKY',.false.,error)
    endif
  endif
  !
  if (.not.allocated(dsky)) then
    allocate (dsky(hclean%gil%dim(1), hclean%gil%dim(2), hclean%gil%dim(3)), stat=ier)
    if (ier.ne.0) then
      call map_message(seve%e,rname,'Memory allocation error')
      error = .true.
      return
    endif
  endif
  !
  call gildas_null(hsky)
  call gdf_copy_header(hclean,hsky,error)
  !
  allocate (dflat(hclean%gil%dim(1),hclean%gil%dim(2)),stat=ier)
  if (ier.ne.0) then
    call map_message(seve%e,rname,'Memory allocation error')
    error = .true.
    return
  endif
  !
  call mos_primary (hsky,dflat,bsize)
  where (dflat.lt.btrunc)
    dflat = 0
  else where
    dflat = 1.0/dflat
  end where
  !
  do i=1,hclean%gil%dim(3)
    dsky(:,:,i) = dclean(:,:,i) * dflat
  enddo
  hsky%loca%addr = locwrd(dsky)
  call gdf_get_extrema(hsky,error)
  hclean%gil%extr_words = def_extr_words
  !
  call sic_def_real ('SKY',dsky,hsky%gil%ndim,hsky%gil%dim,.true.,error)
  save_data(code_save_sky) = .true.
end subroutine primary_comm
!
subroutine mosaic_uvmap(task,line,error)
  use gkernel_interfaces
  use mapping_interfaces, except_this=>mosaic_uvmap
  use clean_def
  use clean_arrays
  use clean_types
  use clean_default
  use gbl_message
  !------------------------------------------------------------------------
  ! @ private
  !
  ! TASK  Compute a Mosaic from a CLIC UV Table with
  ! pointing offset information.
  !
  ! Input :
  !     a UV table with pointing offset information
  !
  ! Ouput
  !   NX NY are the image sizes
  !   NC is the number of channels
  !   NF is the number of different frequencies
  !   NP is the number of pointing centers
  !
  ! 'NAME'.LMV  a 3-d cube containing the uniform noise
  !     combined mosaic, i.e. the sum of the product
  !     of the fields by the primary beam. (NX,NY,NC)
  ! 'NAME'.LOBE the primary beams pseudo-cube (NP,NX,NY,NB)
  ! 'NAME'.WEIGHT the sum of the square of the primary beams (NX,NY,NB)
  ! 'NAME'.BEAM a 4-d cube where each cube contains the synthesised
  !     beam for one field (NX,NY,NB,NP)
  !
  ! All images have the same X,Y sizes
  !------------------------------------------------------------------------
  character(len=*), intent(in) :: task ! Caller (MOSAIC)
  character(len=*), intent(in) :: line ! Command line
  logical, intent(out) :: error
  !
  real(8), parameter :: pi=3.14159265358979323846d0
  real(8), parameter :: f_to_k = 2.d0*pi/299792458.d-6
  character(len=*), parameter :: rname='MOSAIC'
  !
  real, allocatable :: w_mapu(:), w_mapv(:), w_grid(:,:)
  real, allocatable :: res_uv(:,:)
  real(8) new(3)
  real(4) rmega,uvmax,uvmin,uvma
  integer wcol,mcol(2),nfft,sblock
  integer n,ier
  logical one, sorted, shift
  character(len=24) ra_c,dec_c
  character(len=message_length) :: chain
  real cpu0, cpu1
  real(8) :: freq
  real, allocatable :: fft(:)
  integer nx,ny,nu,nv,nc,np
  !
  logical limits
  logical debug
  real ylimn,ylimp
  integer ipen
  !
  real :: thre, btrunc, bsize ! To be initialized
  real :: beamx, beamy
  !
  real, allocatable, target :: dmap(:,:,:), dtmp(:,:,:,:)
  real, allocatable :: dweight(:,:)
  real, allocatable :: dtrunc(:,:)
  real, allocatable :: doff(:,:)
  integer, allocatable :: voff(:)
  real, allocatable :: factorx(:)
  real :: offx, offy, factory, xm, xp, off_range(2)
  integer :: ifield, ic, j, fstart, fend
  integer :: ib, nb, old_ib
  integer, parameter :: o_teles=1
  integer, parameter :: o_trunc=2
  type(projection_t) :: proj
  real(8) :: pos(2)
  !
  debug = sic_present(0,2)
  !
  ! Get beam size from data or command line
  call get_bsize(huv,rname,bsize,btrunc,line,o_teles,o_trunc,error)
  if (error) return
  write(chain,'(a,f10.2,a,f6.0,a)') 'Correcting for a beam size of ',&
    & bsize/pi*180*3600,'" down to ',100*bsize,'% '
  call map_message(seve%i,rname,chain)
  !
  call map_prepare(task,themap,error)
  if (error) return
  !
  one = .true.
  call sic_get_inte('WCOL',wcol,error)
  call sic_get_inte('MCOL[1]',mcol(1),error)
  call sic_get_inte('MCOL[2]',mcol(2),error)
  !
  call sic_get_logi('UV_SHIFT',shift,error)
  if (shift) then
    call sic_get_char('MAP_RA',ra_c,n,error)
    call sic_get_char('MAP_DEC',dec_c,n,error)
    call sic_get_dble('MAP_ANGLE',new(3),error)
  else
    new = 0.d0
  endif
  !
  ! First sort the input UV Table, leaving UV Table in UV_*
  if (shift) then
    call sic_decode(ra_c,new(1),24,error)
    if (error) then
      write(chain,'(A)') 'Input conversion error on phase center'
      call map_message(seve%e,task,chain)
      return
    endif
    call sic_decode(dec_c,new(2),360,error)
    if (error) then
      write(chain,'(A)') 'Input conversion error on phase center'
      call map_message(seve%e,task,chain)
      return
    endif
    new(3) = new(3)*pi/180.0d0
  endif
  call gag_cpu(cpu0)
  !
  ! Note: the sorting should have FIELD ID as primary (slowest varying) key
  ! It needs a mandatory interface for Doffx & Doffy
  ! It may be better to have a separate counting of NP before, as
  ! such a thing would anyway be needed by UV_STAT
  !
  call mosaic_sort (error,sorted,shift,new,uvmax,uvmin, &
    & huv%gil%column_pointer(code_uvt_xoff), &
    & huv%gil%column_pointer(code_uvt_yoff), &
    & np,doff,voff)
  if (error) return
  !
  xm = minval(doff(1,1:np))
  xp = maxval(doff(1,1:np))
  off_range(1) = xp-xm
  xm = minval(doff(2,1:np))
  xp = maxval(doff(2,1:np))
  off_range(2) = xp-xm
  !
  if (.not.sorted) then
    !! Print *,'Done mosaic_sort UV range ',uvmin,uvmax,' sorted ',sorted
    !! call uv_dump_buffers('UV_MOSAIC')
    ! Redefine SIC variables (mandatory)
    call sic_delvariable ('UV',.false.,error)
    call sic_def_real ('UV',duv,huv%gil%ndim,   &
          &      huv%gil%dim,.false.,error)
  else
    !! Print *,'Mosaic was sorted ',uvmin,uvmax,' sorted ',sorted
  endif
  call gag_cpu(cpu1)
  write(chain,102) 'Finished sorting ',cpu1-cpu0
  call map_message(seve%i,task,chain)
  !
  ! This may need to be revised with the Fields Offsets added
  call map_parameters(task,themap,freq,uvmax,uvmin,uvma,off_range)
  !
  themap%xycell = themap%xycell*pi/180.0/3600.0
  !
  ! Get work space, ideally before mapping first image, for
  ! memory contiguity reasons.
  !
  nx = themap%size(1)
  ny = themap%size(2)
  nu = huv%gil%dim(1)
  nv = huv%gil%dim(2)
  nc = huv%gil%nchan
  !
  allocate(w_mapu(nx),w_mapv(ny),w_grid(nx,ny),stat=ier)
  if (ier.ne.0) then
    call map_message(seve%e,task,'Gridding allocation error')
    goto 98
  endif
  !
  do_weig = .true.
  if (do_weig) then
    call map_message(seve%i,task,'Computing weights ')
    if (allocated(g_weight)) deallocate(g_weight)
    if (allocated(g_v)) deallocate(g_v)
    allocate(g_weight(nv),g_v(nv),stat=ier)
    if (ier.ne.0) then
      call map_message(seve%e,task,'Weight allocation error')
      goto 98
    endif
  else
    call map_message(seve%i,task,'Re-using weights')
  endif
  nfft = 2*max(nx,ny)
  allocate(fft(nfft),stat=ier)
  if (ier.ne.0) then
    call map_message(seve%e,task,'FFT allocation error')
    goto 98
  endif
  !
  rmega = 8.0
  ier = sic_ramlog('SPACE_MAPPING',rmega)
  sblock = max(int(256.0*rmega*1024.0)/(nx*ny),1)
  !
  ! New Beam place
  if (allocated(dbeam)) then
    call sic_delvariable ('BEAM',.false.,error)
    deallocate(dbeam)
  endif
  call gildas_null(hbeam)
  !
  ! New dirty image
  if (allocated(ddirty)) then
    call sic_delvariable ('DIRTY',.false.,error)
    deallocate(ddirty)
  endif
  allocate(ddirty(nx,ny,nc),dmap(nx,ny,nc),dtrunc(nx,ny),stat=ier)
  if (ier.ne.0) then
    call map_message(seve%e,task,'Map allocation error')
    goto 98
  endif
  !
  call gildas_null(hdirty)
  hdirty%gil%ndim = 3
  hdirty%gil%dim(1:3) = (/nx,ny,nc/)
  call sic_def_real ('DIRTY',ddirty,hdirty%gil%ndim,   &
       &    hdirty%gil%dim,.true.,error)
  !
  ! Compute the primary beams and weight image
  call gildas_null(hprim)
  if (allocated(dprim)) then
    call sic_delvariable ('PRIMARY',.false.,error)
    deallocate(dprim)
  endif
  if (allocated(dfields)) then
    call sic_delvariable ('FIELDS',.false.,error)
    deallocate(dfields)
  endif
  !
  ! Find out how many beams are required
  call map_beams(task,themap%beam,huv,nx,ny,nb,nc)
  !
  !!Print *,'Done MAP_BEAMS ',themap%beam,nb
  !
  ! An issue is that the map characteristics are defined
  ! in "one_beam". We need to call the proper routines before...
  call map_headers (rname,themap,huv,hbeam,hdirty,hprim,nb,np,mcol)
  !!Print *,'Done MAP_HEADERS '
  hdirty%r3d => ddirty
  !
  ! Define the projection about the Phase center
  call gwcs_projec(huv%gil%a0,huv%gil%d0,-huv%gil%pang,huv%gil%ptyp,proj,error)
  ! POS is here the Offset of the Pointing center relative to the Phase center
  call abs_to_rel (proj,huv%gil%ra,huv%gil%dec,pos(1),pos(2),1)
  !
  if (map_version.lt.0) then
    !
    ! Older MOSAIC code
    call map_message(seve%i,task,'Producing a single beam for all channels')
    !
    hbeam%gil%ndim = 3
    hbeam%gil%dim(1:3)=(/nx,ny,np/)
    hbeam%char%code(3) = 'FIELD'
    hbeam%gil%convert(:,3) = 1.d0
    allocate(dbeam(nx,ny,np,1),stat=ier)
    if (ier.ne.0) then
      call map_message(seve%e,task,'Beam allocation error')
      goto 98
    endif
    call sic_def_real ('BEAM',dbeam,hbeam%gil%ndim,   &
         &    hbeam%gil%dim,.true.,error)
    hbeam%r4d => dbeam
    !
    allocate (factorx(nx), dprim(np,nx,ny,1), dweight(nx,ny), &
      & dtmp(nx,ny,1,1), stat=ier)
    if (ier.ne.0) then
      call map_message(seve%e,task,'Primary beam allocation error')
      goto 98
    endif
    hprim%r4d => dprim
    beamx = hprim%gil%inc(2)/bsize*2.0*sqrt(log(2.0))
    beamy = hprim%gil%inc(3)/bsize*2.0*sqrt(log(2.0))
    !
    ddirty = 0.0
    !
    ! Create the primary beams, lobe and weight images
    do ifield = 1,np
      !
      ! POS is here the Offset of the Pointing center relative to the Phase center
      ! DOFF is the Offset of the Field relative to the Pointing center
      ! So OFFX/Y is the offset of the Field relative to the Phase center in pixel units
      offx = (doff(1,ifield)+pos(1))/hprim%gil%inc(2)
      offy = (doff(2,ifield)+pos(2))/hprim%gil%inc(3)
      !
      do j=1,nx
        factorx(j) = exp(-((j-hprim%gil%ref(2)-offx)*beamx)**2)
      enddo
      do j=1,ny
        factory = exp(-((j-hprim%gil%ref(3)-offy)*beamy)**2)
        dprim(ifield,:,j,1) = factorx(:) * factory
      enddo
    enddo
    deallocate (factorx)
    !
    ! Loop on fields for imaging
    ! Use Dtmp and Dmap as work arrays for beam and image
    hbeam%r3d => dtmp(:,:,:,1)
    hbeam%gil%dim(3) = 1
    hdirty%r3d => dmap
    call map_message(seve%i,task,'Producing a single beam for all channels')
    do ifield = np,1,-1
      !
      do_weig = .true.
      fstart = voff(ifield)      ! Starting Visibility of field
      fend   = voff(ifield+1)-1  ! Ending Visibility of field
      nv = fend-fstart+1
      !
      ! Process sorted UV Table according to the type of beam produced
      call one_beam (task,themap,   &
         &    huv, hbeam, hdirty,   &
         &    nx,ny,nu,nv, duv(:,fstart:fend),   &
         &    w_mapu, w_mapv, w_grid, g_weight, g_v, do_weig,  &
         &    wcol,mcol,fft,   &
         &    sblock,cpu0,error,uvma)
      !
      ! Add it to the "mosaic dirty" image, by multiplying by
      ! the truncated primary beam
      dtrunc = dprim(ifield,:,:,1)
      where(dtrunc.lt.btrunc) dtrunc = 0
      !
      do ic=1,nc
        ddirty(:,:,ic) = ddirty(:,:,ic) + dmap(:,:,ic)*dtrunc(:,:)
      enddo
      !
      ! Save the beam
      dbeam(:,:,ifield,1) = dtmp(:,:,1,1)
    enddo
    !
    ! Reset the good pointers and sizes
    hdirty%r3d => ddirty
    hbeam%r4d => dbeam
    hbeam%gil%ndim = 3
    hbeam%gil%dim(3) = np
    !
    ! Create the Weight image
    !!Print *,'BMIN ',btrunc
    dweight = 0
    call mos_addsq (nx*ny,np,dweight,dprim)
    thre = btrunc**2
    call mos_inverse (nx*ny,dweight,thre)
    !!Print *,'Done MOS_INVERSE'
    !
  else
    !!Print *,'Using untested code, use MAP_BEAM_STEP = -2 if failed'
    !
    ! Code ready but only tested for one Beam for all Frequencies.
    hbeam%gil%ndim = 4
    hbeam%gil%dim(1:4)=(/nx,ny,nb,np/)
    allocate(dbeam(nx,ny,nb,np),stat=ier)
    !
    allocate (dprim(np,nx,ny,nb), dweight(nx,ny), &
      & dtmp(nx,ny,nb,1), factorx(nx), stat=ier)
    if (ier.ne.0) then
      call map_message(seve%e,task,'Primary beam allocation error')
      goto 98
    endif
    hprim%r4d => dprim
    !
    ! Create the primary beams, lobe and weight images
    do ifield = 1,np
      ! bsize is Frequency dependent TBC
      !
      do ib=1,nb
        beamx = hprim%gil%inc(2)/bsize*2.0*sqrt(log(2.0))
        beamy = hprim%gil%inc(3)/bsize*2.0*sqrt(log(2.0))
        offx = (doff(1,ifield)+pos(1))/hprim%gil%inc(2)
        offy = (doff(2,ifield)+pos(2))/hprim%gil%inc(3)
        !
        do j=1,nx
          factorx(j) = exp(-((j-hprim%gil%ref(2)-offx)*beamx)**2)
        enddo
        do j=1,ny
          factory = exp(-((j-hprim%gil%ref(3)-offy)*beamy)**2)
          dprim(ifield,:,j,ib) = factorx(:) * factory
        enddo
      enddo
    enddo
    deallocate (factorx)
    !
    ! Loop on fields for imaging
    ! Use Dtmp and Dmap as work arrays for beam and image
    hbeam%r3d => dtmp(:,:,:,1)
    hbeam%gil%dim(4) = 1
    hdirty%r3d => dmap
    !
    ddirty = 0
    dbeam = 0
    !
    do ifield = np,1,-1
      !
      do_weig = .true.
      fstart = voff(ifield)      ! Starting Visibility of field
      fend   = voff(ifield+1)-1  ! Ending Visibility of field
      nv = fend-fstart+1
      if (debug) then
        Print *,'Ifield ',ifield,fstart,fend
        Print *,'Cols ',wcol,mcol
        Print *,'Sizes ',nx,ny,nu,nv,np,nc
        Print *,'Calling many_beams_para with SBLOCK ',sblock
      endif
      call many_beams_para (task,themap, huv, hbeam, hdirty,   &
         &    nx,ny,nu,nv, duv(:,fstart:fend),   &
         &    g_weight, g_v, do_weig,  &
         &    wcol,mcol,sblock,cpu0,error,uvma)
      !
      old_ib = 0
      !
      do ic=1,nc
        ib = beam_for_channel(ic,hdirty,hbeam)
        if (debug) Print *,'Selected beam ',ib
        ! Add it to the "mosaic dirty" image, by multiplying by
        ! the truncated primary beam
        if (ib.ne.old_ib) then
          dtrunc(:,:) = dprim(ifield,:,:,ib)
          if (debug) Print *,'Set DTRUNC ',ib,' # ',old_ib
          where(dtrunc.lt.btrunc) dtrunc = 0
          old_ib = ib
        endif
        ddirty(:,:,ic) = ddirty(:,:,ic) + dmap(:,:,ic)*dtrunc(:,:)
      enddo
      !
      ! Save the beam - Transposition could be done here if needed
      dbeam(:,:,:,ifield) = dtmp(:,:,:,1)
    enddo
    !
    hbeam%r4d => dbeam
    !
    ! What about the weights here ?
    if (nb.eq.1) then
      dweight = 0
      call mos_addsq (nx*ny,np,dweight,dprim)
      thre = btrunc**2
      call mos_inverse (nx*ny,dweight,thre)
      if (debug) Print *,'Done MOS_INVERSE'
      hbeam%gil%dim(1:4) = [nx,ny,np,1]
      hbeam%gil%convert(:,3) = 1.d0
      hbeam%char%code(3) = 'FIELD'
      hbeam%gil%ndim = 3
    else
      call map_message(seve%e,task,'More than one beam per field - Weights not computed')
    endif
    !
    call sic_def_real ('BEAM',dbeam,hbeam%gil%ndim,   &
         &    hbeam%gil%dim,.true.,error)
  endif
  !
  ! OK we are done (apart from details like Extrema)
  if (.not.error) call map_message(seve%i,task,'Successful completion')
  call sic_def_real ('PRIMARY',dprim,hprim%gil%ndim,   &
   &        hprim%gil%dim,.true.,error)
  save_data(code_save_beam) = .true.
  save_data(code_save_dirty) = .true.
  save_data(code_save_primary) = .true.
  save_data(code_save_fields) = .true.
  !
  call new_dirty_beam
  !!Print *,'Done new_dirty_beam '
  !
  ! Define Min Max
  d_max = hdirty%gil%rmax
  if (hdirty%gil%rmin.eq.0) then
    d_min = -0.03*hdirty%gil%rmax
  else
    d_min = hdirty%gil%rmin
  endif
  !
  error = .false.
  !
  ! Backward compatibility with previous methods
  user_method%trunca = btrunc     ! By convention
  hprim%gil%convert(3,4) = bsize  ! Primary beam size convention
  call sub_mosaic('ON',error)
  !
99 continue
  if (allocated(w_mapu)) deallocate(w_mapu)
  if (allocated(w_mapv)) deallocate(w_mapv)
  if (allocated(w_grid)) deallocate(w_grid)
  if (allocated(fft)) deallocate(fft)
  return
  !
98 call map_message(seve%e,task,'Memory allocation failure')
  error = .true.
  return
  !
102 format(a,f9.2)
end subroutine mosaic_uvmap
!
subroutine map_headers (rname,map,huv,hbeam,hdirty,hprim,nb,nf,mcol)
  use gkernel_interfaces
  use mapping_interfaces, except_this=>map_headers
  use clean_def
  use image_def
  use gbl_message
  !------------------------------------------------------------------------
  ! @ private
  !
  ! MAPPING
  !   Define the image headers
  !------------------------------------------------------------------------
  character(len=*), intent(in) :: rname   ! Calling Task name
  type (uvmap_par), intent(inout) :: map  ! Mapping parameters
  type (gildas), intent(inout) :: huv     ! UV data set
  type (gildas), intent(inout) :: hbeam   ! Dirty beam data set
  type (gildas), intent(inout) :: hdirty  ! Dirty image data set
  type (gildas), intent(inout) :: hprim   ! Primary beam data set
  integer, intent(in) :: nb   ! Number of beams per field
  integer, intent(in) :: nf   ! Number of fields
  integer, intent(inout) :: mcol(2)  ! First and last channel
  ! Global variables:
  !
  real(kind=8), parameter :: clight=299792458d-6 ! Frequency in  MHz
  !
  type(gildas) :: htmp
  integer :: nx   ! X size
  integer :: ny   ! Y size
  integer :: nc   ! Number of channels
  integer :: nd   ! Size of data
  integer icol,lcol,fcol
  real vref,voff,vinc,schunk
  real(kind=4) :: loff,boff
  character(len=message_length) :: chain
  logical :: error
  character(len=4) :: code
  !
  !------------------------------------------------------------------------
  !
  ! Code:
  nx = map%size(1)
  ny = map%size(2)
  !
  vref = huv%gil%ref(1)
  voff = huv%gil%voff
  vinc = huv%gil%vres
  !
  nc = huv%gil%nchan
  if (mcol(1).eq.0) then
    mcol(1) = 1
  else
    mcol(1) = max(1,min(mcol(1),nc))
  endif
  if (mcol(2).eq.0) then
    mcol(2) = nc
  else
    mcol(2) = max(1,min(mcol(2),nc))
  endif
  fcol = min(mcol(1),mcol(2))
  lcol = max(mcol(1),mcol(2))
  nc = lcol-fcol+1
  !
  ! Make beam, not normalized
  call gdf_copy_header(huv,hbeam,error)
  !
  ! Is that right ?
  schunk = nc/nb
  !
  hbeam%gil%ndim = 4
  hbeam%gil%dim(1) = nx
  hbeam%gil%dim(2) = ny
  hbeam%gil%dim(3) = nb
  hbeam%gil%dim(4) = nf
  hbeam%gil%convert(1,1) = nx/2+1
  hbeam%gil%convert(1,2) = ny/2+1
  hbeam%gil%convert(2,1) = 0
  hbeam%gil%convert(2,2) = 0
  hbeam%gil%convert(3,1) = -map%xycell(1)  ! Assume EQUATORIAL system
  hbeam%gil%convert(3,2) = map%xycell(2)
  hbeam%gil%convert(1,3) = (2.d0*(vref-fcol)+schunk+1.d0)/2/schunk ! Correct
  hbeam%gil%convert(2,3) = voff
  hbeam%gil%convert(:,4) = 1.d0
  hbeam%gil%proj_words = 0
  hbeam%gil%extr_words = 0
  hbeam%gil%reso_words = 0
  hbeam%gil%uvda_words = 0
  hbeam%gil%type_gdf = code_gdf_image
  !
  hbeam%char%code(1) = 'ANGLE'
  hbeam%char%code(2) = 'ANGLE'
  hbeam%char%code(3) = 'VELOCITY'
  hbeam%char%code(4) = 'FIELD'
  hbeam%gil%majo = 0.0
  hbeam%loca%size = nx*ny*nb*nf
  !
  ! Prepare the dirty map header
  call gdf_copy_header(hbeam,hdirty,error)
  hdirty%gil%ndim = 3
  hdirty%gil%dim(1) = nx
  hdirty%gil%dim(2) = ny
  hdirty%gil%dim(3) = nc
  hdirty%gil%dim(4) = 1
  hdirty%gil%convert(1,3) = vref-fcol+1
  hdirty%gil%convert(2,3) = voff
  hdirty%gil%convert(3,3) = vinc
  hdirty%gil%proj_words = def_proj_words
  hdirty%gil%uvda_words = 0
  hdirty%gil%type_gdf = code_gdf_image
  hdirty%char%code(1) = 'RA'
  hdirty%char%code(2) = 'DEC'
  hdirty%char%code(3) = 'VELOCITY'
  call equ_to_gal(hdirty%gil%ra,hdirty%gil%dec,0.0,0.0,   &
                  hdirty%gil%epoc,hdirty%gil%lii,hdirty%gil%bii,loff,boff,error)
  if (huv%gil%ptyp.eq.p_none) then
    hdirty%gil%ptyp = p_azimuthal  ! Azimuthal (Sin)
    hdirty%gil%pang = 0.d0     ! Defined in table.
    hdirty%gil%a0 = hdirty%gil%ra
    hdirty%gil%d0 = hdirty%gil%dec
  else
    hdirty%gil%ptyp = p_azimuthal
    hdirty%gil%pang = huv%gil%pang ! Defined in table.
    hdirty%gil%a0 = huv%gil%a0
    hdirty%gil%d0 = huv%gil%d0
  endif
  hdirty%char%syst = 'EQUATORIAL'
  hdirty%gil%xaxi = 1
  hdirty%gil%yaxi = 2
  hdirty%gil%faxi = 3
  hdirty%gil%extr_words = 0          ! extrema not computed
  hdirty%gil%reso_words = 0          ! no beam defined
  hdirty%gil%nois_words = 2
  hdirty%gil%majo = 0
  hdirty%char%unit = 'Jy/beam'
  hdirty%loca%size = nx*ny*nc
  !
  call gildas_null(hprim)
  if (nf.gt.1) then
    call gildas_null(htmp)
    ! Prepare the primary beam cube header
    call gdf_copy_header(hdirty,htmp,error)
    htmp%gil%dim(3) = nf
    htmp%gil%convert(:,3) = 1.d0
    htmp%gil%faxi = 0
    htmp%gil%spec_words = 0
    htmp%char%unit = ' '
    htmp%char%code(3) = 'FIELD'
    code = '312'
    call gdf_transpose_header(htmp,hprim,code,error)
  endif
end subroutine map_headers
!!
subroutine mosaic_sort (error,sorted,shift,new,uvmax,uvmin, &
  & ixoff,iyoff,nf,doff,voff)
  use gkernel_interfaces
  use mapping_interfaces, except_this=>mosaic_sort
  use clean_def
  use clean_arrays
  use gbl_message
  !----------------------------------------------------------------------
  ! @ private-mandatory
  !
  ! MAPPING
  !   Sort the input UV table
  !----------------------------------------------------------------------
  logical, intent(inout) :: sorted      ! Is table sorted ?
  logical, intent(inout) :: shift       ! Do we shift phase center ?
  logical, intent(out) :: error
  real(kind=8), intent(inout) :: new(3) ! New phase center and PA
  real, intent(out) :: uvmin            ! Min baseline
  real, intent(out) :: uvmax            ! Max baseline
  integer, intent(in) :: ixoff, iyoff   ! Offset pointers
  integer, intent(out) :: nf            ! Number of fields
  real, intent(out), allocatable :: doff(:,:)  ! Field offsets
  integer, allocatable :: voff(:)       ! Field visibility offsets
  !
  ! Global variables:
  real(8), parameter :: pi=3.14159265358979323846d0
  real(8), parameter :: f_to_k = 2.d0*pi/299792458.d-6
  real(kind=8) freq, off(3)
  real pos(2), cs(2)
  integer nu,nv,ier
  real, pointer :: duv_previous(:,:), duv_next(:,:)
  !
  ! The UV table is available in HUV%
  if (huv%loca%size.eq.0) then
    call map_message(seve%e,'UV_MAP','No UV data loaded')
    error = .true.
    return
  endif
  nu = huv%gil%dim(1)
  nv = huv%gil%dim(2)
  !
  ! Correct for new phase center if required
  if (shift) then
    if (huv%gil%ptyp.eq.p_none) then
      call map_message(seve%w,'SHIFT','No previous phase center info')
      huv%gil%a0 = huv%gil%ra
      huv%gil%d0 = huv%gil%dec
      huv%gil%pang = 0.d0
      huv%gil%ptyp = p_azimuthal
    elseif (huv%gil%ptyp.ne.p_azimuthal) then
      call map_message(seve%w,'SHIFT','Previous projection type not SIN')
      huv%gil%ptyp = p_azimuthal
    endif
    call uv_shift (new,huv%gil%a0,huv%gil%d0,huv%gil%pang,   &
        &      off,shift)
    huv%gil%posi_words = def_posi_words
    huv%gil%proj_words = def_proj_words
  endif
  !
  sorted = .false.
  if (.not.shift) then
    call check_order_mosaic (duv,nu,nv,ixoff,iyoff,sorted)
  endif
  !
  ! Get center frequency
  freq = gdf_uv_frequency(huv,huv%gil%ref(1))
  !
  if (sorted) then
    !
    ! If already sorted, use it
    call map_message(seve%i,'UV_MOSAIC','UV table is already sorted')
    !
    ! Load Field coordinates and compute UVMAX
    call mosaic_loadfield (duv,nu,nv,ixoff,iyoff,nf,doff,voff,uvmax,uvmin)
  else
    !
    ! Else, create another copy
    call map_message(seve%i,'UV_MOSAIC','Sorting UV table...')
    !
    ! Compute observing frequency, and new phase center in wavelengths
    if (shift) then
      huv%gil%a0 = new(1)
      huv%gil%d0 = new(2)
      huv%gil%pang = new(3)
      cs(1)  =  cos(off(3))
      cs(2)  = -sin(off(3))
      ! Note that the new phase center is counter-rotated because rotations
      ! are applied before phase shift.
      pos(1) = - freq * f_to_k * ( off(1)*cs(1) - off(2)*cs(2) )
      pos(2) = - freq * f_to_k * ( off(2)*cs(1) + off(1)*cs(2) )
    else
      pos(1) = 0.0
      pos(2) = 0.0
      cs(1) = 1.0
      cs(2) = 0.0
    endif
    !
    ! OK, rotate, shift, sort and copy...
    !
    nullify (duv_previous, duv_next)
    !
    call uv_find_buffers ('UV_MOSAIC',nu,nv,duv_previous,duv_next,error)
    if (error) return
    !! call uv_dump_buffers ('UV_MOSAIC - After Find')
    !
    call mosaic_sortuv (nu,nv,huv%gil%ntrail,duv_previous,duv_next,   &
           &        pos,cs,uvmax,uvmin,error,ixoff,iyoff,nf,doff,voff)
    call uv_clean_buffers (duv_previous, duv_next, error)
    if (error) return
    !! call uv_dump_buffers ('UV_MOSAIC- After Clean')
  endif
  !
  ! Now transform UVMAX in kiloWavelength (including 2 pi factor)
  uvmax = uvmax*freq*f_to_k
  uvmin = uvmin*freq*f_to_k
  error = .false.
end subroutine mosaic_sort
!
subroutine mosaic_sortuv (np,nv,ntrail,vin,vout,xy,cs,uvmax,uvmin, &
  & error,ixoff,iyoff,nf,doff,voff)
  use gildas_def
  use gkernel_interfaces
  use mapping_interfaces, except_this=> mosaic_sortuv
  !---------------------------------------------------------------------
  ! @ public-mandatory
  !
  ! MAPPING
  !     Sort a UV table by fields
  !     Rotate, Shift and Sort a UV table for map making
  !     Differential precession should have been applied before.
  !---------------------------------------------------------------------
  integer, intent(in) :: np        ! Size of a visibility
  integer, intent(in) :: nv        ! Number of visibilities
  integer, intent(in) :: ntrail    ! Number of trailing daps
  real, intent(in) :: vin(np,nv)   ! Input visibilities
  real, intent(out) :: vout(np,nv) ! Output visibilities
  real, intent(in) :: xy(2)        ! Phase shift
  real, intent(in) :: cs(2)        ! Frame Rotation
  real, intent(out) :: uvmax       ! Max UV value
  real, intent(out) :: uvmin       ! Min UV value
  integer, intent(in) :: ixoff, iyoff
  integer, intent(out) :: nf
  real, intent(out), allocatable :: doff(:,:)
  integer, intent(out), allocatable :: voff(:)
  logical, intent(out) :: error    !
  ! Local
  real(8), parameter :: pi=3.14159265358979323846d0
  logical, allocatable :: ips(:)       ! Sign of visibility
  real, allocatable :: rpu(:), rpv(:)  ! U,V coordinates
  real, allocatable :: spv(:)          ! Sorted V coordinates
  integer, allocatable :: ipi(:)       ! Index
  real(8), allocatable :: dtr(:)       ! Sorting number
  logical :: sorted
  integer :: ier, ifi, iv
  !
  ! Load U,V coordinates, applying possible rotation (CS),
  ! and making all V negative
  allocate (ips(nv),rpu(nv),rpv(nv),ipi(nv),dtr(nv),stat=ier)
  if (ier.ne.0) then
    error = .true.
    return
  endif
  call loaduv (vin,np,nv,cs,rpu,rpv,ips,uvmax,uvmin)
  !!Print *,'UVMIN ',uvmin,' UVMAX ',uvmax,' NP ',np,' NV ',nv
  !
  ! Identify number of fields
  call loadfiuv (vin,np,nv,dtr,ipi,sorted,ixoff,iyoff,rpv,nf,doff)
  !
  ! Sort by fields (major number) then V (fractionary part)
  if (.not.sorted) then
    !!Print *,'Sorting UV data '
    call gr8_trie (dtr,ipi,nv,error)
    deallocate (dtr,stat=ier)
    allocate (spv(nv),stat=ier)
    !
    ! One must sort RPV here to use SORTUV later...
    do iv=1,nv
      spv(iv) = rpv(ipi(iv))
    enddo
    rpv(:) = spv(:)
    deallocate (spv,stat=ier)
  else
    deallocate (dtr,stat=ier)
    Print *,'UV Data is already sorted '
  endif
  !!Read(5,*) ifi
  !
  ! Apply phase shift and copy to output visibilities
  call sortuv (vin,vout,np,nv,ntrail,xy,rpu,rpv,ips,ipi)
  !
  ! Compute start & end of fields
  allocate(voff(nf+1),stat=ier)
  ifi = 1
  voff(ifi) = 1
  do iv=1,nv
    if ( (doff(1,ifi).ne.vout(ixoff,iv)) .or. &
      &  (doff(2,ifi).ne.vout(iyoff,iv)) ) then
      ifi = ifi+1
      voff(ifi) = iv
    endif
  enddo
  voff(nf+1) = nv+1
  !
  !!Print *,'XOFF ',ixoff,' YOFF ',iyoff
  do ifi=1,min(nf,20)
    write(*,'(I4,A,2F10.2,2I10)') ifi,' DOFF ', &
      & doff(1,ifi)*180.*3600./pi, &
      & doff(2,ifi)*180.*3600./pi, &
      & voff(ifi), voff(ifi+1)-1
  enddo
  !
  error = .false.
  !
end subroutine mosaic_sortuv
!
subroutine check_order_mosaic(visi,np,nv,ixoff,iyoff,sorted)
  !----------------------------------------------------------
  ! @ private
  !
  ! MAPPING
  !   Check if visibilites are sorted.
  !   Chksuv does a similar job, but using V values and an index.
  !----------------------------------------------------------
  integer, intent(in) :: np       ! Size of a visibility
  integer, intent(in) :: nv       ! Number of visibilities
  real, intent(in) :: visi(np,nv) ! Visibilities
  integer, intent(in) :: ixoff    ! X pointing column
  integer, intent(in) :: iyoff    ! Y pointing column
  logical, intent(out) :: sorted
  !
  real vmax,xoff,yoff
  integer iv
  !
  vmax = visi(2,1)
  xoff = visi(ixoff,1)
  yoff = visi(iyoff,1)
  !
  do iv=2,nv
    if (visi(2,iv).lt.vmax) then
      if (visi(ixoff,iv).eq.xoff .and. visi(iyoff,iv).eq.yoff) then
        !!Print *,'Unsorted V at ',iv,visi(2,iv),vmax
        sorted = .false.
        return
      endif
      ! else, this is a new offset
      xoff = visi(ixoff,iv)
      yoff = visi(iyoff,iv)
    else if (visi(ixoff,iv).eq.xoff .and. visi(iyoff,iv).eq.yoff) then
      ! ok, things progress normally
      continue
    else
      ! Unsorted offset
      !!Print *,'Unsorted Position offset at ',iv
      sorted = .false.
      return
    endif
    vmax = visi(2,iv)
  enddo
  sorted = .true.
end subroutine check_order_mosaic
!
subroutine loadfiuv (visi,np,nv,dtr,it,sorted,ixoff,iyoff,rpv,nf,doff)
  !----------------------------------------------------------------------
  ! @ public-mandatory
  !
  ! MAP    UVSORT routines
  !     Load field numbers into work arrays for sorting.
  !----------------------------------------------------------------------
  integer, intent(in)  :: np                   ! Size of a visibility
  integer, intent(in)  :: nv                   ! Number of visibilities
  real, intent(in) :: visi(np,nv)              ! Input visibilities
  real(8), intent(out) :: dtr(nv)              ! Output field number
  integer, intent(out) :: it(nv)               ! Indexes
  logical, intent(out) :: sorted               !
  integer, intent(in)  :: ixoff                ! X pointer
  integer, intent(in)  :: iyoff                ! Y pointer
  real(4), intent(in)  :: rpv(nv)              ! V Values
  integer, intent(out) :: nf                   ! Number of fields
  real(kind=4), intent(out), allocatable :: doff(:,:)  ! Fields offsets
  !
  integer :: iv, itime, iant, jant
  integer :: ifi, mfi, kfi, nfi, ier
  real(8) :: vmax, d0
  real(kind=4), allocatable :: dtmp(:,:)
  !
  ! Scan how many fields
  nfi = 1
  mfi = 100
  !
  ! V are negative values, so this 1 + max(abs(V))
  vmax = 1.0d0-minval(rpv)
  !
  allocate(dtmp(2,mfi),stat=ier)
  dtmp(1,1) = visi(ixoff,1)
  dtmp(2,1) = visi(iyoff,1)
  dtr(1) = 1.d0+rpv(1)/vmax ! We have here 0 =< dtr < 1
  !
  do iv=2,nv
    kfi = 0
    do ifi=1,nfi
      if (visi(ixoff,iv).eq.dtmp(1,ifi) .and. &
      & visi(iyoff,iv).eq.dtmp(2,ifi) ) then
        dtr(iv) = dble(ifi)+rpv(iv)/vmax
        kfi = ifi
        exit
      endif
    enddo
    !
    ! New field
    if (kfi.eq.0) then
      if (nfi.eq.mfi) then
        allocate(doff(2,mfi),stat=ier)
        doff(:,:) = dtmp(:,:)
        deallocate(dtmp)
        allocate(dtmp(2,2*mfi),stat=ier)
        dtmp(:,1:mfi) = doff
        deallocate(doff)
        mfi = 2*mfi
      endif
      nfi = nfi+1
      dtmp(1,nfi) = visi(ixoff,iv)
      dtmp(2,nfi) = visi(iyoff,iv)
      dtr(iv) = dble(nfi)+rpv(iv)/vmax   ! nfi-1 =< dtr < nfi
    endif
  enddo
  !
  nf = nfi
  allocate(doff(2,nfi),stat=ier)
  doff(1:2,:) = dtmp(1:2,1:nfi)
  !
  do iv=1,nv
    it(iv) = iv
  enddo
  !
  ! DTR must in the end be ordered and increasing.
  vmax = dtr(1)
  do iv = 1,nv
    if (dtr(iv).lt.vmax) then
      sorted = .false.
      return
    endif
    vmax = dtr(iv)
  enddo
  sorted = .true.
  !
end subroutine loadfiuv





