subroutine uvstat_map(task,freq,map,uvmax,uvmin,print)
  use gkernel_interfaces
  use mapping_interfaces, only : map_message
  use clean_def
  use clean_arrays
  use clean_types
  use clean_default
  use gbl_message
  !----------------------------------------------------------------------
  ! @ private
  !
  ! Estimate (and Display) the MAP parameters which will be used for
  ! UV_MAP, MX or UV_RESTORE
  !-----------------------------------------------------------------------
  character(len=*), intent(in)    :: task ! Input task name (UV_MAP or MX)
  type (uvmap_par), intent(inout) :: map  ! Map parameters
  real(8), intent(inout) :: freq          ! Observing frequency
  real(4), intent(in) :: uvmax, uvmin     ! Min & Max UV in m
  logical, optional :: print              ! Printout results
  !
  real(8), parameter :: pi=3.14159265358979323846d0
  real(8), parameter :: f_to_k = 2.d0*pi/299792458.d-6
  real, parameter :: pixel_per_beam=5.0
  character(len=80) :: chain
  real :: hppb, diam, uvmi, uvma
  logical do_print
  !
  do_print = present(print)
  !
  freq = gdf_uv_frequency(huv,huv%gil%ref(1))
  if (do_print) then
    write(chain,'(A,F12.6,A)') "Observed rest frequency       ",freq*1d-3," GHz"
    call map_message(seve%i,task,chain)
  endif
  !
  ! Primary beam
  if (huv%gil%majo.ne.0) then
    hppb = huv%gil%majo
  else
    diam = 0.
    if (huv%gil%nteles.ne.0) then
      diam = huv%gil%teles(1)%diam
    endif
    if (diam.eq.0) diam = 15.0
    hppb = 63.0*pi/180.0/3600.0*(12.0/diam)*(100.e3/freq) ! [rad] Primary beam size
  endif
  hppb = hppb*180*3600/pi
  if (do_print) then
    write(chain,'(A,F12.1,A)') "Half power primary beam       ",hppb," arcsec"
    call map_message(seve%i,task,chain)
  endif
  !
  ! say "   Phase center RA and Dec       "'&1%int%cent%sex[1]'" "'&1%int%cent%sex[2]'
  !
  uvma = uvmax/(freq*f_to_k)
  uvmi = uvmin/(freq*f_to_k)
  if (do_print) then
    write(chain,100) huv%gil%dim(2),huv%gil%nchan
    call map_message(seve%i,task,chain)
    write(chain,101) uvmi,uvma,' meters'
    call map_message(seve%i,task,chain)
    write(chain,101) uvmin*1e-3/2.0/pi,uvmax*1e-3/2.0/pi,' kiloWavelength'
    call map_message(seve%i,task,chain)
  endif
  !
  ! Now get MAP_CELL and MAP_SIZE
  map%xycell(1) = 0.01*nint(360.0*3600.0*100.0/uvmax/pixel_per_beam)
  if (map%xycell(1).le.0.02) map%xycell(1) =   &
         &        0.002*nint(360.0*3600.0*500.0/uvmax/pixel_per_beam)
  map%xycell(2) = map%xycell(1)  ! In ", rounded to 0.02 or 0.002"
  !
  ! Get a wide enough field of view...
  !
  map%size = 0
  if (map%field(1).ne.0) then
    if (map%field(2).eq.0) map%field(2) = map%field(1)
  else
    ! The factor 2 on the primary beam size is to go down to 6.25 % level by default
    map%field(1) = 2*hppb
    ! The factor 2 on uvmax/uvmin is to allow all scales down to uvmin
    map%field(2) = 2*pixel_per_beam*uvmax/uvmin * map%xycell(1)
    ! The Min is to avoid going to very large images when uvmin is close to 0
    map%field = min(map%field(1),map%field(2))
  endif
  !
  map%size = map%field/map%xycell !
  !
  ! Round to power of two : perhaps not the best option here.
  ! Could become too small or uselessly too large, and be compensated by a different
  ! cell size choice.
  !
  map%size =   2**nint(log(real(map%size))/log(2.0))
  !
  if (do_print) then
    write(chain,'(A,F12.1,a,F12.1,A)') 'Field of view / Largest Scale ',map%field(1),' x ',map%field(2),' "'
    call map_message(seve%i,task,chain)
    write(chain,'(A,F11.1,A,F11.1,A)') 'Image size ',map%size(1)*map%xycell(1),' x ',map%size(2)*map%xycell(2),'" '
    call map_message(seve%i,task,chain)
    !
    write(chain,103) map%size
    call map_message(seve%i,task,chain)
    write(chain,104) map%xycell
    call map_message(seve%i,task,chain)
    !
  endif
  !
100 format('Found ',i12,' Visibilities, ',i4,' channels')
101 format('Baselines ',f9.1,' - ',f9.1,a)
103 format('Recommended map size is   ',i4,' by ',i4)
104 format('Recommended pixel size is ',f8.3,' by ',f8.3,'"')
end subroutine uvstat_map
!
subroutine uv_stat_comm(line,error)
  use gkernel_interfaces
  use mapping_interfaces, except_this=>uv_stat_comm
  use clean_def
  use clean_default
  use clean_arrays
  use code_names
  use gbl_message
  !------------------------------------------------------------------------
  ! @ private
  !
  ! MAPPING   Support for command UV_STAT
  !       Analyse a UV data set to define approximate beam size,
  !       field of view, expected "best beam", etc...
  ! Input :
  !     a precessed UV table
  ! Output :
  !     a precessed, rotated, shifted UV table, sorted in V,
  !     ordered in (U,V,W,D,T,iant,jant,nchan(real,imag,weig))
  !     a beam image ?
  !------------------------------------------------------------------------
  character(len=*), intent(inout) :: line
  logical, intent(out) :: error
  !
  real(kind=8), parameter :: pi=3.14159265358979323846d0
  real(kind=8), parameter :: f_to_k = 2.d0*pi/299792458.d-6
  !
  integer, parameter :: mt=100
  integer datelist(mt)
  !
  integer i
  integer ndates
  character(len=8) mode, argum
  character(len=4) :: wei
  real, allocatable :: beams(:,:), fft(:)
  real uv_taper(3),map_cell(2),uniform(2),uvmax,uvmin,uvm
  real start,step,taper,result(10,8)
  integer map_size(2),wcol,ctype,n,mcol(2),ier,nu,nv,noff
  integer(kind=index_length) :: dim(4)
  logical sorted, shift
  real(kind=8) :: freq, new(3)
  real(kind=4) :: map_field(2)
  real(4), save :: bmax, bmin
  logical, save :: first = .true.
  logical :: print
  character(len=message_length) :: chain
  character(len=7) :: rname = 'UV_STAT'
  character(len=8) :: smode(5)
  data smode /'CELL','HEADER','SETUP','TAPER','WEIGHT'/
  !
  if (sic_present(0,1)) then
    call sic_ke(line,0,1,argum,n,.false.,error)
    if (error) return
    call sic_ambigs (rname,argum,mode,n,smode,5,error)
    if (error) return
  else
    mode = 'ALL'
  endif
  !
  freq = gdf_uv_frequency(huv)
  !
  select case (mode)
  case ('CELL','TAPER','WEIGHT')
    !
    start = 0.0
    call sic_r4(line,0,2,start,.false.,error)
    start = abs(start)
    step = 0.0
    call sic_r4(line,0,3,step,.false.,error)
    step = abs(step)
    !
    ctype = 1
    call sic_get_real('UV_TAPER[1]',uv_taper(1),error)
    if (error) return
    call sic_get_real('UV_TAPER[2]',uv_taper(2),error)
    call sic_get_real('UV_TAPER[3]',uv_taper(3),error)
    call sic_get_real('UV_CELL[1]',uniform(1),error)
    call sic_get_real('UV_CELL[2]',uniform(2),error)
    call sic_get_inte('WCOL',wcol,error)
    call sic_get_inte('MCOL[1]',mcol(1),error)
    call sic_get_inte('MCOL[2]',mcol(2),error)
    !
    call get_weightmode(rname,wei,error)
    if (error) return
    !
    ! First sort the input UV Table, leaving UV Table in UV_*
    shift = .false.
    call uv_sort (error,sorted,shift,new,uvmax,uvmin)
    if (error) return
  case ('ALL','HEADER','SETUP')
    ! For HEADERS, do the minimum work for speed...
    ! The UV table is available in HUV%
    if (huv%loca%size.eq.0) then
      call map_message(seve%e,rname,'No UV data loaded')
      error = .true.
      return
    endif
    call uvgmax (huv,duv,uvmax,uvmin)
    ! Now transform UVMAX in kiloWavelength (including 2 pi factor)
    uvmax = uvmax*freq*f_to_k
    uvmin = uvmin*freq*f_to_k
    error = .false.
  end select
  !
  ! New code: For the time being, this is somewhat inconsistent
  ! with the use of uv_sort which uses a different frequency
  !
  bmax = uvmax/(freq*f_to_k)
  bmin = uvmin/(freq*f_to_k)
  if (first) then
!    call sic_defstructure('UVSTAT',.true.,error)
!    call sic_def_real('UVSTAT%BMAX',bmax,0,dim,.false.,error)
!    call sic_def_real('UVSTAT%BMIN',bmin,0,dim,.false.,error)
    call sic_def_real('UV_BMAX',bmax,0,dim,.false.,error)
    call sic_def_real('UV_BMIN',bmin,0,dim,.false.,error)
    first = .false.
  endif
  !
  ! That's all for HEADER
  select case (mode)
  case ('HEADER')
    call gdf_print_header(huv)
    call uv_listheader(huv,duv,mt,datelist,ndates,freq)
    call uv_printoffset('HEADER',themap)
  case ('SETUP')
    call map_copy_par(default_map,themap)
    call uvstat_map(rname,freq,themap,uvmax,uvmin,print)
    error = .false.
  case ('ALL')
    call gdf_print_header(huv)
    call uv_listheader(huv,duv,mt,datelist,ndates,freq)
    call uv_printoffset('HEADER',themap)
    call map_copy_par(default_map,themap)
    call uvstat_map(rname,freq,themap,uvmax,uvmin,print)
  case default
    !
    ! back to usual code
    uvm = uvmax/(freq*f_to_k)
    !
    if (mode.eq.'TAPER') then
      if (step.eq.0.) step = sqrt(2.0)
      if (start.eq.0.) start = 10*nint(uvm/160.0)
      taper = 16*start
      if (wei.eq.'NATU') then
        uniform(1) = 0.
        uniform(2) = 0.
      endif
    else
      taper = sqrt(uv_taper(1)*uv_taper(2))
      if (taper.ne.0) then
        taper = min(taper,uvm)
      else
        taper = uvm
      endif
    endif
    !
    ! Define MAP_CELL and MAP_SIZE
    taper = uvm*freq*f_to_k
    map_cell(1) = 0.02*nint(180.0*3600.0*50.0/taper/4.0)
    if (map_cell(1).le.0.02) then
      map_cell(1) = 0.002*nint(180.0*3600.0*500.0/taper/4.0)
    endif
    map_cell(2) = map_cell(1)    ! In ", rounded to 0.01"
    if (mode.eq.'TAPER') then
      map_size(1) = 256
      map_size(2) = 256
    else
      map_size(1) = 64
      map_size(2) = 64
    endif
    !
    write(chain,102) uvm,uvmax*1e-3/2.0/pi
    call map_message(seve%i,rname,chain)
    write(chain,103) map_size
    call map_message(seve%i,rname,chain)
    write(chain,104) map_cell
    call map_message(seve%i,rname,chain)
    !
    ! Redefine some parameters
    map_cell(1) = map_cell(1)*pi/180.0/3600.0   ! In radians
    map_cell(2) = map_cell(2)*pi/180.0/3600.0
    !
    ! Process sorted UV Table according to the type of beam produced
    n = 2*max(map_size(1),map_size(2))
    allocate (beams(map_size(1),map_size(2)),fft(n),stat=ier)
    if (ier.ne.0) then
      call map_message(seve%e,rname,'Memory allocation failure')
      error = .true.
      return
    endif
    call uniform_beam (' ',uv_taper,   &
       &    map_size,map_cell,uniform,wcol,mcol,fft,   &
       &    error,mode,beams,1,start,step,uvm,result,huv,duv)
    deallocate (beams,fft)
    !
    call sic_delvariable('BEAM_SHAPE',.false.,error)
    dim(1) = 10
    dim(2) =  8
    call sic_def_real('BEAM_SHAPE',result,2,dim,.true.,error)
    error = .false.
  end select
  !
  102   format('Maximum baseline is   ',f8.1,' m,  ',f8.1,' kWavelength')
  103   format('Map size is   ',i4,' by ',i4)
  104   format('Pixel size is ',f6.3,' by ',f5.3,'"')
end subroutine uv_stat_comm
!
subroutine uv_printoffset(rname,map)
  use clean_def
  use gbl_message
  use mapping_interfaces, only : map_message
  ! @ private
  character(len=*), intent(in) :: rname
  type(uvmap_par), intent(in) :: map
  !
  real(8), parameter :: pi=3.14159265358979323846d0
  real(4), parameter :: rad_to_sec = 180.0*3600.0/pi
  integer :: n,j
  character(len=60) :: chain
  !
  if (map%nfields.eq.0) return
  !
  n = abs(map%nfields)
  if (map%nfields.lt.0) then
    write(chain,'(I5,A)') n, ' Phase offsets in mosaic'
  else
    write(chain,'(I5,A)') n, ' Pointing offsets in mosaic'
  endif
  call map_message(seve%i,rname,chain)
  do j=1,n-1,2
    write(*,'(2(A,F12.2,A,F12.2,A))') &
    &   '(',map%offxy(1,j)*rad_to_sec,' ,',map%offxy(2,j)*rad_to_sec,' )', &
    &   '(',map%offxy(1,j+1)*rad_to_sec,' ,',map%offxy(2,j+1)*rad_to_sec,' )'
  enddo
  if (mod(n,2).ne.0)     write(*,'(A,F12.2,A,F12.2,A)') &
    &   '(',map%offxy(1,n)*rad_to_sec,' ,',map%offxy(2,n)*rad_to_sec,' )'
end subroutine uv_printoffset


