#include <math.h>
#include <stdio.h>
#include "wipe.h"
#ifndef  M_PI
#define  M_PI   3.14159265358979323846
#endif

/******************************************************************************/
/******************************************************************************/
void     wipe_aperture(
int      *P_nx,    /* number of pixels in the x direction (input)             */
int      *P_ny,    /* number of pixels in the y direction (input)             */
int      *P_nf,    /* number of points in the exper. list Le+ (input)         */
float    *P_xA,    /* multiplicative factor for major axis of Hr (input)      */
float    *P_xB,    /* multiplicative factor for minor axis of Hr (input)      */
float    *P_wipe_major,  /* major axis of the ellipse HR (output)             */
float    *P_wipe_minor,  /* minor axis of the ellipse HR (output)             */
float    *P_wipe_angle,  /* angle of the ellipse HR (output)                  */
float    *P_ang,   /* rotation angle to apply to default aperture (input)     */
float    *P_Eta,   /* mesure of the cross interpolation to perform (input)    */
float    *Vu,      /* U coordinate of the uv plane (input)                    */
float    *Vv,      /* V coordinate of the uv plane (input)                    */
float    *Vn,      /* Sigma on visibility (1|sqrt(2) (input)                  */
int      *He,      /* Experimental aperture (output)                          */
float    *HeW,     /* Weighted experimental aperture (output)                 */
int      *Hr,      /* Synthesized aperture (output)                           */
int      *Dr)      /* Effective support of the Neat Beam (output)             */
/******************************************************************************/
/******************************************************************************/
{
register int i, j, k;
int          nx = *P_nx, ny = *P_ny, nf = *P_nf;
int          nxs2 = nx/2, nys2 = ny/2;
float        xA = *P_xA, xB = *P_xB, Eta = *P_Eta;
double       u, v, x, y, surf_Hr, surf_Vr, delta;
double       J_uu, J_vv, J_uv, a11, a22, a12, A, B;
double       Lp, Lm, Vp[2], Vm[2], dist, distc;
float        ang=*P_ang, alpha1, alpha2, alpha3, alpha4;

/******* computing experimental aperture He *********************************
*******/
for (i=0; i<nx*ny; i++) He[i] = Hr[i] = Dr[i] = 0;
for (i=0; i<nx*ny; i++) HeW[i] = 0.0;

for (k=0; k<nf; k++)
  {
  u = (double)Vu[k]; u *= ((double) nx);
  v = (double)Vv[k]; v *= ((double) ny);
  i = nys2 + ((int) v);
  j = nxs2 + ((int) u);
  He[i*ny+j] += 1;
  HeW[i*ny+j] += Vn[k];
  i = nys2 - ((int) v);
  j = nxs2 - ((int) u);
  He[i*ny+j] += 1;
  HeW[i*ny+j] += Vn[k];
  }

/******* computing inertial moments J_uu, J_vv, and J_uv of He *****************
*******/
J_uu = J_vv = J_uv = surf_Hr = 0.0;
for (i=0; i<nx; i++)
  {
  v = ((double) (i - nxs2));
  for (j=0; j<ny; j++)
    {
    u = ((double) (j - nys2));
/*
    if (He[i*ny+j] >= 1)
*/
    if ((He[i*ny+j] >= 1) && (HeW[i*ny+j] != 0))
       {
/*
       J_uu += u*u*He[i*ny+j];
       J_vv += v*v*He[i*ny+j];
       J_uv += u*v*He[i*ny+j];
*/
       J_uu += u*u;
       J_vv += v*v;
       J_uv += u*v;
       surf_Hr += 1;
       }
    }
  }
J_uu *= (2.0*M_PI*M_PI);        /*       / J_uu  J_uv \       */
J_vv *= (2.0*M_PI*M_PI);        /*   J = |            |       */
J_uv *= (2.0*M_PI*M_PI);        /*       \ J_uv  J_vv /       */

/******* vp & VP eigenvalues and eigenvectors of the inertial moments matrix *******
*******/
delta = (J_uu-J_vv)*(J_uu-J_vv) + 4.0*J_uv*J_uv;
Lp = 0.5*((J_uu+J_vv) + sqrt(delta));
Lm = 0.5*((J_uu+J_vv) - sqrt(delta)); /* Lm < Lp */
Vp[1] = (Lp-J_uu)/J_uv; Vp[0] = 1.0; a11 = sqrt(Vp[0]*Vp[0]+Vp[1]*Vp[1]);
Vm[1] = (Lm-J_uu)/J_uv; Vm[0] = 1.0; a22 = sqrt(Vm[0]*Vm[0]+Vm[1]*Vm[1]);
Vp[0] /= a11; Vp[1] /= a11;
Vm[0] /= a22; Vm[1] /= a22;

a11 = surf_Hr/(M_PI*sqrt(Lm*Lp));
Lp *= a11; Lm *= a11;
Lp = 1.0/Lp; Lm = 1.0/Lm;
Lp /= (double)(xA*xA); Lm /= (double)(xB*xB);
/*
   Modifying the major,minor and rotation angle of the
   default ellipse corresponding to the aperture to synthesize
*/
ang = ang*M_PI/180.0;
alpha1 = cos(ang)*Vp[0]+sin(ang)*Vp[1];
alpha2 = -sin(ang)*Vp[0]+cos(ang)*Vp[1];
alpha3 = cos(ang)*Vm[0]+sin(ang)*Vm[1];
alpha4 = -sin(ang)*Vm[0]+cos(ang)*Vm[1];
Vp[0] = alpha1;
Vp[1] = alpha2;
Vm[0] = alpha3;
Vm[1] = alpha4;

a11 = Lp*Vp[0]*Vp[0]+Lm*Vm[0]*Vm[0];
a22 = Lp*Vp[1]*Vp[1]+Lm*Vm[1]*Vm[1];
a12 = Lp*Vp[0]*Vp[1]+Lm*Vm[0]*Vm[1];

/******* computing the aperture to be synthesized Hr ********************************
*******/
for (i=0; i<nx; i++)
  {
  v = ((double) (i - nxs2));
  for (j=0; j<ny; j++)
    {
    u = ((double) (j - nys2));
    if ((a11*u*u+a22*v*v+2.0*a12*u*v) <= 1.0) Hr[i*ny+j] = 1;
    else Hr[i*ny+j] = 0;
    }
  }

*P_wipe_major = (float)(1.0/sqrt(Lp));
*P_wipe_minor = (float)(1.0/sqrt(Lm));

dist = 0.0;
for (i=0; i<nx; i++)
for (j=0; j<ny; j++)
  if (Hr[i*ny+j] == 1.0)
     {
     u = ((double) (j - nys2));
     v = ((double) (i - nxs2));
     distc = sqrt(u*u+v*v);
     if (distc > dist)
        {
        dist = distc;
        if (u != 0.0)
           *P_wipe_angle = atan(v/u);
        else
           *P_wipe_angle = 0.0;
        }
     }
*P_wipe_angle = (float)(*P_wipe_angle*180.0/M_PI);

/******* computing inertial moments J_uu, J_vv, and J_uv of Hr ****************
*******/
J_uu = J_vv = J_uv = surf_Hr = 0.0;
for (i=0; i<nx; i++)
  {
  v = ((double) (i - nxs2));
  for (j=0; j<ny; j++)
    {
    u = ((double) (j - nys2));
    J_uu += u*u*Hr[i*ny+j];
    J_vv += v*v*Hr[i*ny+j];
    J_uv += u*v*Hr[i*ny+j];
    surf_Hr += Hr[i*ny+j];
    }
  }
J_uu *= (2.0*M_PI*M_PI);        /*       / J_uu  J_uv \       */
J_vv *= (2.0*M_PI*M_PI);        /*   J = |            |       */
J_uv *= (2.0*M_PI*M_PI);        /*       \ J_uv  J_vv /       */
surf_Hr /= ((double) (nx*ny));

/*******  eigenvalues of the inertial moments matrix *********************
*******/
delta = (J_uu-J_vv)*(J_uu-J_vv) + 4.0*J_uv*J_uv;
Lp = 0.5*((J_uu+J_vv) + sqrt(delta));
Lm = 0.5*((J_uu+J_vv) - sqrt(delta)); /* Lm < Lp */

/******* computing the parameters of the ellipsoid of resolution Dr ****************
*******/
A = 1.0/sqrt(Lp); /* minor axis */
B = 1.0/sqrt(Lm); /* major axis */
surf_Vr = M_PI*A*B;
delta = Eta*Eta/(surf_Hr*surf_Vr);
a11 = J_uu/delta;
a22 = J_vv/delta;
a12 = J_uv/delta;

/******* computing the ellipsoid of resolution Dr *******************************
*******/
for (i=0; i<nx; i++)
  {
  y = ((double) (i - nxs2));
  for (j=0; j<ny; j++)
    {
    x = ((double) (j - nys2));
    if ((a11*x*x+a22*y*y+2.0*a12*x*y) <= 1.0) Dr[i*ny+j] = 1;
    else Dr[i*ny+j] = 0;
    }
  }
}
