#include  <stdio.h>
#include  <math.h>

#ifndef  M_PI
#define  M_PI   3.14159265358979323846
#endif

/*******************************************************************************
@
@ FUNCTION......: ffs1D_taper
@
@ PURPOSE.......: Computes the Fourier transform (tapering function) of the
@                 gaussian interpolation function used for Fast Fourier
@                 Sampling.
@
@ INPUT.........: col = number of columns
@                 l = number of terms in the Fourier serie
@                 r,a = coefficients of the gaussian interpolation function
@                       >> rect[sqrt(u^2+v^2)/r]*exp[-a^2.(u^2+v^2)] <<
@
@ OUTPUT........: tape[0..col-1] = float vector
@
@ VERSION.......: June 1994
@
@ AUTHOR........: Eric ANTERRIEU
@
*******************************************************************************/
/******************************************************************************/
void ffs1D_taper( float *tape, int col, int r, float a, int l)
/******************************************************************************/
{
register int       i, k;
float    u, x;

for (i=0; i<col; i++)
  {
  x = -0.5 + ((float) i)/((float) col);
  tape[i] = 0.0;
  for (k=1; k<=((l*r)/2); k++)
    {
    u = ((float) k)/((float) l);
    tape[i] += (exp(-a*a*u*u)*cos(2.0*M_PI*u*x));
    }
  tape[i] *= (2.0/((float) l));
  tape[i] += (1.0/((float) l));
  }
}

/*******************************************************************************
@
@ FUNCTION......: inv_ffs2D_real
@
@ PURPOSE.......: Computes Fourier components on a rectangular grid from complex
@                 visibilities along the upper half of the uv coverage.
@
@ INPUT.........: lin,col = number of lines and number of columns
@                 nbr_vis = number of complex visibilities
@                 vis_u[0..nbr_vis-1],vis_v[0..nbr_vis-1] = Fourier frequencies
@                 vis_r[0..nbr_vis-1],vis_i[0..nbr_vis-1] = complex visibilities
@                 r,a = coefficients of the gaussian interpolation function
@                       >> rect[sqrt(u^2+v^2)/r]*exp[-a^2.(u^2+v^2)] <<
@
@ OUTPUT........: mat[0..lin-1][0..col-1] = float FFT2D matrix
@
@ RETURN........: 0 if OK
@                -1 if Fourier frequencies out of range (outside ]-0.5;+0.5[)
@
@ VERSION.......: June 1994
@
@ AUTHOR........: Eric ANTERRIEU
@
*******************************************************************************/

/******************************************************************************/
int inv_ffs2D_real(
                   float *mat, int lin, int col, float *vis_r, float *vis_i,
                   float *vis_u, float *vis_v, int nbr_vis, int r, float a)
/******************************************************************************/
{
register int      i, j, k;
int      flag = 0, lins2, v1, v2, vi, cols2, u1, u2, uj;
float   re, im, imm, u, uu, v, vv, arg;

lins2 = lin/2; cols2 = col/2;

k = 0;
while (k < nbr_vis)
  {
  if  (vis_v[k] >  0.5) flag = -1;
  if  (vis_v[k] < -0.5) flag = -1;
  if  (vis_u[k] >  0.5) flag = -1;
  if  (vis_u[k] < -0.5) flag = -1;
  if ((vis_u[k] <  0.0) && (vis_v[k] == 0.0)) flag = -1;
  k++;
  if (flag == -1) k = nbr_vis;
  }
/* fprintf(stderr,"flag = %d\n",flag); */
if (flag != 0) return(-1);

for (i=0; i<lin; i++)
for (j=0; j<col; j++) mat[lin*i+j] = 0.0;
for (k=0; k<nbr_vis; k++)
  {
  if (vis_r != NULL) re = vis_r[k]; else re = 1.0;
  if (vis_i != NULL) imm = vis_i[k]; else imm = 0.0;
  u = ((float) col)*vis_u[k]; u1 = ((int) ceil(u-r)); u2 = ((int) floor(u+r));
  v = ((float) lin)*vis_v[k]; v1 = ((int) ceil(v-r)); v2 = ((int) floor(v+r));

  for (vi=v1; vi<=v2; vi++)
  for (uj=u1; uj<=u2; uj++)
    {
    i = vi; vv = 0.5*(v - ((float) vi)); im = imm;
    j = uj; uu = 0.5*(u - ((float) uj));
    arg = 0.25*exp(-a*a*(uu*uu + vv*vv));

    if (vi < 0) {j = -j;  i = -i;  im = -im;}
    if ((vi == 0) && (uj < 0)) {j = -j;  im = -im;}
    if (uj == (-cols2)) {j = -j;  im = -im;}
    if ((vi == (lins2)) && (uj < 0)) {j = -j;  im = -im;}

    if ((i == 0) || (i == lins2))
      {
      if ((j == 0) || (j == cols2))
        {
        mat[lin*i+j] += 2.0*re*arg;
        }
      else
        {
        mat[lin*i+j] += re*arg;
        mat[lin*i+col-j] += im*arg;
        }
      }
    else
      {
      if (j >= 0)
        {
        mat[lin*i+j] += re*arg;
        mat[lin*(lin-i)+j] += im*arg;
        }
      else
        {
        mat[lin*i+col+j] += re*arg;
        mat[lin*(lin-i)+col+j] += im*arg;
        }
      }
    }
  }

return(0);
}
