program uv_gain
  use gildas_def
  use gkernel_interfaces
  use image_def
  use gbl_format
  use gbl_message
  !---------------------------------------------------------------------
  ! GILDAS
  ! Compute Gains by comparing the UVT table and
  !       a UV model file.
  ! Input : a UV table (observed)
  !         a UV model
  ! Output: a UV table, with the gains.
  ! Output: a UV calibrated table
  !---------------------------------------------------------------------
  ! Local
  type(gildas) :: out,y,z
  integer, parameter :: mant=256  ! Valid for ALMA...
  integer, parameter :: mbas=mant*(mant-1)/2
  character(len=*), parameter :: rname='UV_GAIN'
  logical :: error, do_amp, do_pha, antenna_based
  integer(4) ::  n, ichan, nvis, ncol, ier
  character(len=filename_length) :: uvdata, uvmodel, uvgain, uvcal, &
    & uvsol, name
  real :: tinteg, snr
  integer :: refant
  !
  integer, allocatable :: index(:)
  real(8), allocatable :: itime(:)
  real, allocatable :: work(:,:)
  integer :: ntimes, nantes
  !
  ! Code
  call gildas_open
  call gildas_char('UVDATA$',uvdata)
  call gildas_char('UVMODEL$',uvmodel)
  call gildas_char('UVGAIN$',uvgain)
  call gildas_char('UVCAL$',uvcal)
  call gildas_real('TINTEG$',tinteg,1)
  call gildas_real('SNR$',snr,1)
  call gildas_inte('CHANNEL$',ichan,1)
  call gildas_logi('DO_AMP$',do_amp,1)
  call gildas_logi('DO_PHA$',do_pha,1)
  call gildas_logi('ANTENNA_BASED$',antenna_based,1)
  call gildas_inte('REFERENCE_A$',refant,1)
  if (antenna_based) call gildas_char('UVSOL$',uvsol)
  call gildas_close
  !
  if (uvdata.eq.uvcal) then
    call map_message(seve%f,rname, &
    & 'Input and Self-calibrated table must differ.')
    error = .true.
    goto 999
  endif
  !
  ! Input file : data, in Z slot
  n = len_trim(uvdata)
  if (n.le.0) goto 999
  name = uvdata(1:n)
  call gildas_null(z, type = 'UVT')
  call gdf_read_gildas(z, name, '.uvt', error)
  if (error) then
    call map_message(seve%e,rname,'Cannot read input UV table ')
    goto 999
  endif
  !
  ! Input file : model, in Y slot
  n = len_trim(uvmodel)
  if (n.le.0) goto 999
  call gildas_null(y, type = 'UVT')
  name = uvmodel(1:n)
  call gdf_read_gildas(y, name, '.uvt', error)
  if (error) then
    call map_message(seve%e,rname,'Cannot read model UV table ')
    goto 999
  endif
  !
  ! Prepare output gain table : in X slot
  call gildas_null(out, type = 'UVT')
  call gdf_copy_header (y, out, error)
  n = len_trim(uvgain)
  if (n.eq.0) goto 999
  name  = uvgain(1:n)
  call sic_parsef(name,out%file,' ','.uvt')
  call map_message(seve%i,rname,'Creating UV table '//trim(out%file))
  call gdf_create_image (out, error)
  if (.not.error) call gdf_allocate (out,error)
  if (error) then
    call map_message(seve%e,rname,'Cannot create Gain Table')
    goto 999
  endif
  !
  nvis = z%gil%dim(2)
  ncol = z%gil%dim(1)
  !
  allocate (index(nvis), itime(nvis), stat=ier)
  if (antenna_based) then
    allocate(work(10,nvis),stat=ier)
  endif
  !
  ! Get baseline-based gains ...
  ! they go into X slot (data in Z, model in Y)
  call do_b_gain(do_amp,do_pha,nvis,ncol,ichan,itime,   &
     &    index,z%r2d,y%r2d,out%r2d)
  !!Print *,'Done do_b_gain'
  !
  ! Model table is not needed any more.
  deallocate (y%r2d, stat=ier)
  call gdf_close_image (y, error)
  !
  ! Compute antenna-based gains ...
  ! they go in Work array
  ! then overwrite the baseline-based array in OUT%R2D
  if (antenna_based) then
    call do_a_gain(do_amp,do_pha,nvis,itime,tinteg,snr,   &
     &      out%r2d,work, refant, ntimes, nantes)
    call r4tor4(work,out%r2d,nvis*10)
    if (refant.eq.0) refant = 1
    call do_p_gain(ntimes,nantes,uvsol,refant,error)
  endif
  !
  ! Prepare output calibrated table
  call gdf_copy_header(z, y, error)
  n = len_trim(uvcal)
  if (n.eq.0) goto 999
  name  = uvcal(1:n)
  call sic_parsef(name,y%file,' ','.uvt')
  call map_message(seve%i,rname,'Creating UV table '//trim(y%file))
  call gdf_create_image (y,error)
  if (.not.error) call gdf_allocate (y, error)
  if (error) then
    call map_message(seve%e,rname,'Cannot create Calibrated Table')
    goto 999
  endif
  !
  ! Fill in calibrated table
  call do_cal(ncol,z%gil%nchan,nvis,z%r2d,y%r2d,out%r2d,index)
  !
  call gdf_write_data (y, y%r2d, error)
  call gdf_write_data (out, out%r2d, error)
  call map_message(seve%i,rname,'Successful completion')
  call sysexi(1)
  !
  999   call sysexi (fatale)
end program uv_gain
!
subroutine do_cal(ncol,nch,nvis,data,cal,gain,index)
  !---------------------------------------------------------------------
  ! Global:
  ! Dummy:
  !---------------------------------------------------------------------
  integer, intent(in) :: ncol                   ! Number of column in visi
  integer, intent(in) :: nvis                   ! Number of visibilities
  integer, intent(in) :: nch                    ! Number of channels
  real, intent(in) :: data(ncol,nvis)           ! Raw data
  real, intent(out) :: cal(ncol,nvis)           ! Calibrated data
  real, intent(in) :: gain(10,nvis)             ! Gain array
  integer, intent(in) :: index(nvis)            ! Visibility order
  ! Local
  ! Local:
  integer :: iv, jv, k
  complex :: zdata, zgain, zcal
  !-----------------------------------------------------------------------
  do iv=1, nvis
    jv = index(iv)
    do k=1, 7
      cal(k,jv) = data(k,jv)
    enddo
    zgain = cmplx(gain(8,iv),gain(9,iv))
    if (gain(10,iv).lt.0) then
      zgain = 0
    endif
    do k=8, 5+3*nch, 3
      if (zgain.ne.0) then
        zdata = cmplx(data(k,jv),data(k+1,jv))
        zcal = zdata / zgain
        cal(k,jv) = real(zcal)
        cal(k+1,jv) = aimag(zcal)
        cal(k+2,jv) = data(k+2,jv)*abs(zgain)**2
      else
        cal(k,jv) = data(k,jv)
        cal(k+1,jv) = data(k+1,jv)
        cal(k+2,jv) = data(k+2,jv)
      endif
    enddo
    if (8+3*nch.le.ncol) then
      cal(8+3*nch:ncol,iv) = data(8+3*nch:ncol,iv)
    endif
  enddo
end subroutine do_cal
!
subroutine do_a_gain(do_amp,do_pha,nvis,times,tinteg,snr,gain,work, &
  & refant,ntimes,nantes)
  use gildas_def
  use gbl_message
  !---------------------------------------------------------------------
  ! Compute complex gains (antenna-based solution)
  !---------------------------------------------------------------------
  logical, intent(in)  :: do_amp                 ! Calibrate amplitude ?
  logical, intent(in)  :: do_pha                 ! Compute phase ?
  integer, intent(in)  :: nvis                   ! Number of visibilities
  real(8), intent(in)  :: times(nvis)            ! Times of visibilities
  real, intent(in)  :: tinteg                    ! Integration time
  real, intent(in)  :: snr                       ! snr parameter
  real, intent(in)  :: gain(10,nvis)             ! Input baseline gain array
  real, intent(inout)  :: work(10,nvis)          ! Output antenna gain array
  integer, intent(in) :: refant                  ! Reference antenna
  integer, intent(out) :: ntimes                 ! Number of time steps
  integer, intent(out) :: nantes                 ! Number of antennas
  ! Local
  character(len=*), parameter :: rname='UV_GAIN'
  integer, parameter :: mant=256  ! ALMA uses stations rather than antennas
  integer, parameter :: mbas=mant*(mant-1)/2
  integer ::  iref, iv, jv, ia, ja, ib, i, nant, nbas, j, k, nant_old
  integer ::  la(mant), iant(mbas), jant(mbas), lia, lja, igood, isol, iflag
  real :: ampli, phase, ua(mant), va(mant), ug, vg, d2, dmin
  real :: aa(mant),wa(mant),ampsnr(mant),maxsnr,minsnr
  real(8) ::  x(mbas), y(mbas), amp(mbas), pha(mbas)
  real(8) ::  w(mbas), ca(mant), cp(mant), wp(mant)
  real(8) ::  wk2(mant,mant), wk3(mant), ss(mbas), t, told
  logical :: error, ok
  character(len=256) :: mess
  !-----------------------------------------------------------------------
  nbas = 0
  igood = 0
  isol = 0
  do j=2, mant
    do i=1, j-1
      nbas = nbas+1
      iant(nbas) = i
      jant(nbas) = j
    enddo
  enddo
  !
  ! The algorithm will find a better value for ref antenna,
  ! the one neares to the array center.
  iref = refant
  !
  ! Loop on visibilities
  !
  told = -1d10
  ntimes = 0
  nantes = 0
  do iv=1, nvis
    t = times(iv)
    if (told.ne.t) then
      ntimes = ntimes+1
      !Print *,'New output time ',iv,told,t,ntimes
      told = t
      nant_old = 0
      jv = iv
      do while (jv.gt.1 .and. t .ge. told-tinteg/2)
        jv = jv - 1
        t = times(jv)
      enddo
      nant = 0
      do i=1, mant
        la(i) = 0
        ca(i) = 0.0
        cp(i) = 0.0
      enddo
      do i=1, mbas
        x(i) = 0
        y(i) = 0
        w(i) = 0
      enddo
      !
      ! Search how many antennas at this time, and average
      ! over all visibilities which fit the time range
      !
      ok = (jv.lt.nvis .and. t.le.told+tinteg/2.)
      do while (ok)
        !
        ! Build the antenna pointer list
        ia = nint(gain(6,jv))
        if (la(ia).eq.0) then
          nant = nant+1
          la(ia) = nant
        endif
        ja = nint(gain(7,jv))
        if (la(ja).eq.0) then
          nant = nant+1
          la(ja) = nant
        endif
        !
!        if (t.eq.told) then
!          nant_old = nant
!        else if (nant.ne.nant_old) then
!          Print *,'Number of antenna changed at visibility ',jv,nant,nant_old
!        endif
        ! In the end, we use this formula...
        !    ia = la(nint(gain(6,iv)))
        !    ja = la(nint(gain(7,iv)))
        ! which figures out which sequence number is each antenna
        !
        ! Here we set in UA,VA the "Antenna" U,V coordinate
        ! relative to Antenna #1
        lia = la(ia)
        lja = la(ja)
        if (lia.eq.1) then
          ua(lja) = gain(1,jv)
          va(lja) = gain(2,jv)
        elseif (lja.eq.1) then
          ua(lia) = -gain(1,jv)
          va(lia) = -gain(2,jv)
        endif
        !
        ! Here we average the visibility into the X,Y,W buffer
        ! we also set start antenna < end antenna
        if (lia.lt.lja) then
          ib = (lja-1)*(lja-2)/2+lia
          y(ib) = y(ib) + gain(9,jv)*gain(10,jv)
        else
          ib = (lia-1)*(lia-2)/2+lja
          y(ib) = y(ib) - gain(9,jv)*gain(10,jv)
        endif
        x(ib) = x(ib) + gain(8,jv)*gain(10,jv)
        w(ib) = w(ib) + gain(10,jv)
        jv = jv + 1
        t = times(jv)
        ok = (jv.lt.nvis .and. t.le.told+tinteg/2.)
      enddo
      !
      ! get the reference antenna, by locating the one
      ! with the shortest baselines
      !
      nbas = nant*(nant-1)/2
      !
      ! Compute the centroid of antenna positions (still relative to Antenna #1)
      ug = 0
      vg = 0
      do ia=1, nant
        ug = ug+ua(ia)
        vg = vg+va(ia)
      enddo
      ug = ug/nant
      vg = vg/nant
      dmin = 1e10
      !
      ! Locate the antenna closest to centroid (this is independent of absolute pos.)
      iref = 1
      do ia=1, nant
        d2 = (ua(ia)-ug)**2 + (va(ia)-vg)**2
        if (d2.lt.dmin) then
          dmin = d2
          iref = ia
        endif
      enddo
      !
      ! This is ambiguous if antenna renumbering occured
      if (refant.ne.0) iref = refant
      !
      ! Normalize the time averaged visibilities
      do ib=1, nbas
        if (w(ib).ne.0) then
          x(ib) = x(ib)/w(ib)
          y(ib) = y(ib)/w(ib)
        endif
      enddo
      !
      ! amplitudes
      if (do_amp) then
        do ib=1, nbas
          if (w(ib).ne.0) then
            amp(ib) = log(x(ib)**2+y(ib)**2)/2.
          endif
        enddo
        call gain_ant(1, nbas, iant, jant, iref,   &
     &          nant, amp, w, wk2, wk3, ss, ca,wp, error)
      endif
      !
      ! phases
      if (do_pha) then
        ! Solve for amplitude first to estimate S/N ratio
        do ib=1, nbas
          if (w(ib).ne.0) then
            amp(ib) = log(x(ib)**2+y(ib)**2)/2.
            w(ib)   = w(ib)*(x(ib)**2+y(ib)**2)
          endif
        enddo
        call gain_ant(1, nbas, iant, jant, iref,   &
     &          nant, amp, w, wk2, wk3, ss, ca,wp, error)
        do ia = 1, nant
          isol = isol + 1
          aa(ia) = exp(2*ca(ia))
          wa(ia) = wp(ia)/aa(ia)**2
          ampsnr(ia) = aa(ia)*sqrt(wa(ia))*1e3
          !
          ! Reset amplitude to 1 now
          ca(ia) = 0
        enddo
        maxsnr = maxval(ampsnr(1:nant))
        minsnr = minval(ampsnr(1:nant))
        if (maxsnr.gt.snr) then
          do ib=1, nbas
            if (w(ib).ne.0) then
              pha(ib) = atan2(y(ib),x(ib))
            endif
          enddo
          call gain_ant(2, nbas, iant, jant, iref,   &
     &            nant, pha, w, wk2, wk3, ss, cp, wp, error)
          iflag = 0
          do ia =1, nant
            !
            ! Keep only antenna solution with required snr
            if (ampsnr(ia).lt.snr) then
              cp(ia) = 0
              iflag = iflag+1
            else
              igood = igood + 1
            endif
          enddo
        else
          iflag = nant
        endif
        if (iflag.ne.0) then
          write(mess,101) told,maxsnr,minsnr,iflag
        else
          write(mess,101) told,maxsnr,minsnr
        endif
101   format('Time ',F10.1,' SNR: Max ',F5.1,' Min ',F5.1,I3,' antennas ignored')
        call map_message(seve%i,rname,mess)
      endif
      !
      ! Write the Phase on Unit 2, Amplitude on Unit 3
      write(2,*) iref,told,nant
      write(2,100) cp(1:nant)*180.0/3.14159
100   format (8(2x,f6.1))
      nantes = max(nantes,nant)
    endif
    !
    ! Fill in the antenna gains in "work"
    do k=1,7
      work(k,iv) = gain(k,iv)
    enddo
    ia = la(nint(gain(6,iv)))
    ja = la(nint(gain(7,iv)))
    ampli = 1./exp(ca(ia)+ca(ja)) ! Not sure for this one...
    phase = - cp(ia) + cp(ja)     ! This is THE solution
    !! Print *,'IV ',ampli, 180*phase/3.14159
    work(8,iv) = ampli * cos(phase)
    work(9,iv) = ampli * sin(phase)
    !
    ! Flag data with no solution
    if (.not.error) then
       work(10,iv) = gain(10,iv)
    else
       work(10,iv) = -abs(gain(10,iv))
    endif
  enddo
  write(mess,'(A,I0,A,I0,A)') "Found ",igood,"/",isol," good solutions"
  call map_message(seve%i,rname,mess)
end subroutine do_a_gain
!
subroutine do_b_gain(do_amp,do_pha,nvis,ncol,icol,times,index,   &
     &    data,model,gain)
  !---------------------------------------------------------------------
  ! we use channel ic from data table
  ! model and gain are assumed to be continuum tables.
  ! gain will contain the observed baseline gains.
  !---------------------------------------------------------------------
  logical, intent(in) :: do_amp                 ! Calibrate amplitude
  logical, intent(in) :: do_pha                 ! Calibrate phase
  integer, intent(in) :: nvis                   ! Number of visibilities
  integer, intent(in) :: ncol                   ! Size of a visibility
  integer, intent(in) :: icol                   ! Reference column
  real(8), intent(out) :: times(nvis)           ! Time stamps of visibilities
  integer, intent(out) :: index(nvis)           ! Index of visibilities
  real, intent(in) :: data(ncol,nvis)           ! Raw visibilities
  real, intent(in) :: model(10,nvis)            ! Modeled visibilities
  real, intent(out) :: gain(10,nvis)            ! Gain solution
  ! Local
  logical :: error
  ! Local:
  integer :: iv, i, jv, k
  complex :: zgain, zdata, zmodel
  real :: wgain, wdata, again, date0
  !-----------------------------------------------------------------------
  !
  ! Get the chronological order:
  date0 = data(4,1)
  do iv=1, nvis
    times(iv) = (data(4,iv)-date0)*86400.d0+data(5,iv)  ! Beware of Real(8) precision
  enddo
  call gr8_trie (times,index,nvis,error)
  !
  ! Sorted data in gain:
  do iv=1,nvis
    jv = index(iv)
    do i=1, 7
      gain(i,iv) = data(i,jv)
    enddo
    k = (icol-1)*3+7
    zdata = cmplx(data(k+1,jv),data(k+2,jv))
    wdata = data(k+3,jv)
    zmodel = cmplx(model(8,jv),model(9,jv))
    if (zmodel.ne.0) then
      zgain = zdata/zmodel
      wgain = wdata*abs(zmodel)**2
      if (.not.do_amp) then
        again = abs(zmodel)
        if (again.ne.0) then
          zgain = zgain*again
        else
          zgain = 1.0
        endif
        wgain = wdata
      elseif (.not.do_pha) then
        zgain = abs(zgain)
      endif
    else
      zgain = 0
      wgain = 0
    endif
    gain(8,iv) = real(zgain)
    gain(9,iv) = aimag(zgain)
    gain(10,iv) = wgain
  enddo
end subroutine do_b_gain
!
subroutine gain_ant(iy,nbas,iant,jant,iref,nant,y,w,wk2,wk3,ss,c,wc,error)
  use gkernel_interfaces
  !---------------------------------------------------------------------
  !     gain_ant computes a weighted least-squares approximation
  !     to the antenna amplitudes or phases
  !     for an arbitrary set of baseline data points.
  !---------------------------------------------------------------------
  integer, intent(in) :: iy                ! 1 for log(amplitude), 2 for phase
  integer, intent(in) :: nbas              ! the number of baselines
  integer, intent(in) :: iant(nbas)        ! start antenna for each baseline
  integer, intent(in) :: jant(nbas)        ! end  antenna for each baseline
  integer, intent(in) :: iref              ! Reference antenna for phases
  integer, intent(in) :: nant              ! the number of antennas
  real(8), intent(in) :: y(nbas)           ! the data values
  real(8), intent(in) :: w(nbas)           ! weights
  real(8), intent(inout) :: wk2(nant,nant) ! work space
  real(8), intent(inout) :: wk3(nant)      ! work space
  real(8), intent(out) :: ss(nbas)         ! rms of fit for each baseline
  real(8), intent(out) :: c(nant)          ! the gains
  real(8), intent(out) :: wc(nant)         ! the weights
  logical :: error                  !
  ! Global
  integer :: zant
  ! Local
  real(8) :: norm, tol
  !
  parameter (tol=1d-14)
  integer :: i, ia, ib, nantm1, ja, l, iter, ier
  real(8) ::  wi, ww, yi
  !------------------------------------------------------------------------
  ! Code
  !
  ss = 0.
  !
  ! Check that the weights are positive.
  !
  error = .false.
  do ib = 1, nbas
    if (w(ib).lt.0.0d0) then
      call gagout('E-GAIN_ANT, Weights not positive')
      error = .true.
      return
    endif
  enddo
  !
  ! Amplitude case is simple...
  !
  if (iy.eq.1) then
    wk2 = 0.0d0
    wk3 = 0.0d0
    !
    ! store the upper-triangular part of the normal equations in wk2
    ! and the right-hand side in wk3.
    do ib=1,nbas
      wi = w(ib)
      if (wi.gt.0) then
        ia = iant(ib)
        ja = jant(ib)
        wk2(ia,ia) = wk2(ia,ia)+wi
        wk2(ia,ja) = wk2(ia,ja)+wi
        wk2(ja,ia) = wk2(ja,ia)+wi
        wk2(ja,ja) = wk2(ja,ja)+wi
      endif
    enddo
    do ib=1, nbas
      ia = iant(ib)
      ja = jant(ib)
      wi = w(ib)*y(ib)
      wk3(ia) = wk3(ia) + wi
      wk3(ja) = wk3(ja) + wi
    enddo
    !
    ! Solve the system of normal equations by first computing the Cholesky
    ! factorization
    call mth_dpotrf ('GAIN_ANT','U',nant,wk2,nant,error)
    if (error) return
    call mth_dpotrs ('GAIN_ANT','U',nant,1,wk2,nant,wk3,nant,ier)
    if (ier.ne.0) return
    do i=1,nant
      c(i) = wk3(i)
      wc(i) = wk2(i,i)**2
    enddo
    !
    ! Phase is more complicated ...
  elseif (iy.eq.2) then
    nantm1 = nant-1
    do i=1,nant
      c(i) = 0.0d0
      wc(i) = 0.0d0
    enddo
    !
    ! start iterating
    norm = 1e10
    iter = 0
    do while (norm.gt.tol .and. iter.lt.100)
      iter = iter + 1
      do i=1,nantm1
        do l=1,nantm1
          wk2(l,i) = 0.0d0
        enddo
        wk3(i) = 0.0d0
      enddo
      do ib=1,nbas
        wi = w(ib)
        if (wi.gt.0) then
          ia = zant(iant(ib),iref)
          ja = zant(jant(ib),iref)
          if (ia.ne.0) then
            wk2(ia,ia) = wk2(ia,ia)+wi
          endif
          if (ja.ne.0) then
            wk2(ja,ja) = wk2(ja,ja)+wi
          endif
          if (ia.ne.0 .and. ja.ne.0) then
            wk2(ja,ia) = wk2(ja,ia)-wi
            wk2(ia,ja) = wk2(ia,ja)-wi
          endif
        endif
      enddo
      do ib=1, nbas
        if (w(ib).gt.0) then
          yi = y(ib)
          ia = iant(ib)
          ja = jant(ib)
          yi = yi+(c(ia)-c(ja))
        else
          yi = 0
        endif
        yi = sin(yi)
        ia = zant(iant(ib),iref)
        ja = zant(jant(ib),iref)
        wi = w(ib)*yi
        if (ia.ne.0) then
          wk3(ia) = wk3(ia) - wi
        endif
        if (ja.ne.0) then
          wk3(ja) = wk3(ja) + wi
        endif
      enddo
      !
      ! Solve the system of normal equations by first computing the Cholesky
      ! factorization
      call mth_dpotrf('GAIN_ANT','U',nantm1,wk2,nant,error)
      if (error) return
      call mth_dpotrs ('GAIN_ANT','U',nantm1,1,wk2,nant,   &
     &        wk3,nantm1,ier)
      if (ier.ne.0) return
      !  Add the result to c:
      norm = 0
      do ia=1,nant
        i = zant(ia,iref)
        if (i.ne.0) then
          ww = wk3(i)
          c(ia) = c(ia)+ww
          wc(ia) = wk2(i,i)**2
          norm = norm+ww**2
        endif
      enddo
    enddo
  endif
  !
  ! I would need to set SS, the rms on each baseline. How can it be
  ! computed ?
end subroutine gain_ant
!
function zant(i,r)
  integer :: zant                   ! intent(out)
  integer, intent(in) :: i          !
  integer, intent(in) :: r          !
  if (i.eq.r) then
    zant = 0
  elseif (i.gt.r) then
    zant = i-1
  else
    zant = i
  endif
  return
end function zant
!*
!=========================================================================
! Linear Algebra: use LAPACK routines
!
subroutine mth_dpotrf (name, uplo, n, a, lda, error)
  character(len=*) :: name           !
  character(len=*) :: uplo           !
  integer :: n                       !
  integer :: lda                     !
  real(8) :: a(lda,*)                 !
  logical :: error                   !
  ! Local
  integer :: info
  !
  !  Purpose
  !  =======
  !
  !  DPOTRF computes the Cholesky factorization of a real symmetric
  !  positive definite matrix A.
  !
  !  The factorization has the form
  !     A = U**T * U,  if UPLO = 'U', or
  !     A = L  * L**T,  if UPLO = 'L',
  !  where U is an upper triangular matrix and L is lower triangular.
  !
  !  This is the block version of the algorithm, calling Level 3 BLAS.
  !
  !  Arguments
  !  =========
  !
  !  UPLO    (input) CHARACTER*1
  !          = 'U':  Upper triangle of A is stored;
  !          = 'L':  Lower triangle of A is stored.
  !
  !  N       (input) INTEGER
  !          The order of the matrix A.  N >= 0.
  !
  !  A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
  !          On entry, the symmetric matrix A.  If UPLO = 'U', the leading
  !          N-by-N upper triangular part of A contains the upper
  !          triangular part of the matrix A, and the strictly lower
  !          triangular part of A is not referenced.  If UPLO = 'L', the
  !          leading N-by-N lower triangular part of A contains the lower
  !          triangular part of the matrix A, and the strictly upper
  !          triangular part of A is not referenced.
  !
  !          On exit, if INFO = 0, the factor U or L from the Cholesky
  !          factorization A = U**T*U or A = L*L**T.
  !
  !  LDA     (input) INTEGER
  !          The leading dimension of the array A.  LDA >= max(1,N).
  !
  !  INFO    (output) INTEGER
  !          = 0:  successful exit
  !          < 0:  if INFO = -i, the i-th argument had an illegal value
  !          > 0:  if INFO = i, the leading minor of order i is not
  !                positive definite, and the factorization could not be
  !                completed.
  !
  ! Call LAPACK routine
  call dpotrf  (uplo, n, a, lda, info )
  call mth_fail(name,'MTH_DPOTRF',info,error)
end subroutine mth_dpotrf
!
subroutine mth_dpotrs (name,   &
     &    uplo, n, nrhs, a, lda, b, ldb, info )
  character(len=*) :: name           !
  character(len=*) :: uplo           !
  integer :: n                       !
  integer :: nrhs                    !
  integer :: lda                     !
  real(8) :: a(lda,*)                 !
  integer :: ldb                     !
  real(8) :: b(ldb,*)                 !
  ! Local
  integer :: info
  logical :: error
  !
  !  Purpose
  !  =======
  !
  !  DPOTRS solves a system of linear equations A*X = B with a symmetric
  !  positive definite matrix A using the Cholesky factorization
  !  A = U**T*U or A = L*L**T computed by DPOTRF.
  !
  !  Arguments
  !  =========
  !
  !  UPLO    (input) CHARACTER*1
  !          = 'U':  Upper triangle of A is stored;
  !          = 'L':  Lower triangle of A is stored.
  !
  !  N       (input) INTEGER
  !          The order of the matrix A.  N >= 0.
  !
  !  NRHS    (input) INTEGER
  !          The number of right hand sides, i.e., the number of columns
  !          of the matrix B.  NRHS >= 0.
  !
  !  A       (input) DOUBLE PRECISION array, dimension (LDA,N)
  !          The triangular factor U or L from the Cholesky factorization
  !          A = U**T*U or A = L*L**T, as computed by DPOTRF.
  !
  !  LDA     (input) INTEGER
  !          The leading dimension of the array A.  LDA >= max(1,N).
  !
  !  B       (input/output) DOUBLE PRECISION array, dimension (LDB,NRHS)
  !          On entry, the right hand side matrix B.
  !          On exit, the solution matrix X.
  !
  !  LDB     (input) INTEGER
  !          The leading dimension of the array B.  LDB >= max(1,N).
  !
  !  INFO    (output) INTEGER
  !          = 0:  successful exit
  !          < 0:  if INFO = -i, the i-th argument had an illegal value
  !
  ! Call LAPACK routine
  call dpotrs  (uplo, n, nrhs, a, lda, b, ldb, info )
  call mth_fail(name,'MTH_DPOTRF',info,error)
end subroutine mth_dpotrs
!
subroutine mth_dpbtrf (name, uplo, n, kd, ab, ldab, error)
  character(len=*) :: name             !
  character(len=*) :: uplo             !
  integer :: n                         !
  integer :: kd                        !
  integer :: ldab                      !
  real(8) :: ab(ldab,*)                 !
  logical :: error                     !
  ! Local
  integer :: info
  !
  !  Purpose
  !  =======
  !
  !  DPBTRF computes the Cholesky factorization of a real symmetric
  !  positive definite band matrix A.
  !
  !  The factorization has the form
  !     A = U**T * U,  if UPLO = 'U', or
  !     A = L  * L**T,  if UPLO = 'L',
  !  where U is an upper triangular matrix and L is lower triangular.
  !
  !  Arguments
  !  =========
  !
  !  UPLO    (input) CHARACTER*1
  !          = 'U':  Upper triangle of A is stored;
  !          = 'L':  Lower triangle of A is stored.
  !
  !  N       (input) INTEGER
  !          The order of the matrix A.  N >= 0.
  !
  !  KD      (input) INTEGER
  !          The number of superdiagonals of the matrix A if UPLO = 'U',
  !          or the number of subdiagonals if UPLO = 'L'.  KD >= 0.
  !
  !  AB      (input/output) DOUBLE PRECISION array, dimension (LDAB,N)
  !          On entry, the upper or lower triangle of the symmetric band
  !          matrix A, stored in the first KD+1 rows of the array.  The
  !          j-th column of A is stored in the j-th column of the array AB
  !          as follows:
  !          if UPLO = 'U', AB(kd+1+i-j,j) = A(i,j) for max(1,j-kd)<=i<=j;
  !          if UPLO = 'L', AB(1+i-j,j)    = A(i,j) for j<=i<=min(n,j+kd).
  !
  !          On exit, if INFO = 0, the triangular factor U or L from the
  !          Cholesky factorization A = U**T*U or A = L*L**T of the band
  !          matrix A, in the same storage format as A.
  !
  !  LDAB    (input) INTEGER
  !          The leading dimension of the array AB.  LDAB >= KD+1.
  !
  !  INFO    (output) INTEGER
  !          = 0:  successful exit
  !          < 0:  if INFO = -i, the i-th argument had an illegal value
  !          > 0:  if INFO = i, the leading minor of order i is not
  !                positive definite, and the factorization could not be
  !                completed.
  !
  !  Further Details
  !  ===============
  !
  !  The band storage scheme is illustrated by the following example, when
  !  N = 6, KD = 2, and UPLO = 'U':
  !
  !  On entry:                       On exit:
  !
  !      *    *   a13  a24  a35  a46      *    *   u13  u24  u35  u46
  !      *   a12  a23  a34  a45  a56      *   u12  u23  u34  u45  u56
  !     a11  a22  a33  a44  a55  a66     u11  u22  u33  u44  u55  u66
  !
  !  Similarly, if UPLO = 'L' the format of A is as follows:
  !
  !  On entry:                       On exit:
  !
  !     a11  a22  a33  a44  a55  a66     l11  l22  l33  l44  l55  l66
  !     a21  a32  a43  a54  a65   *      l21  l32  l43  l54  l65   *
  !     a31  a42  a53  a64   *    *      l31  l42  l53  l64   *    *
  !
  !  Array elements marked * are not used by the routine.
  !
  ! Call LAPACK routine
  call dpbtrf  (uplo, n, kd, ab, ldab, info )
  call mth_fail(name,'MTH_DPBTRF',info,error)
end subroutine mth_dpbtrf
!
subroutine mth_dpbtrs (name,   &
     &    uplo, n, kd, nrhs, ab, ldab, b, ldb, error)
  character(len=*) :: name             !
  character(len=*) :: uplo             !
  integer :: n                         !
  integer :: kd                        !
  integer :: nrhs                      !
  integer :: ldab                      !
  real(8) :: ab(ldab,*)                 !
  integer :: ldb                       !
  real(8) :: b(ldb,*)                   !
  logical :: error                     !
  ! Local
  integer :: info
  !
  !  Purpose
  !  =======
  !
  !  DPBTRS solves a system of linear equations A*X = B with a symmetric
  !  positive definite band matrix A using the Cholesky factorization
  !  A = U**T*U or A = L*L**T computed by DPBTRF.
  !
  !  Arguments
  !  =========
  !
  !  UPLO    (input) CHARACTER*1
  !          = 'U':  Upper triangular factor stored in AB;
  !          = 'L':  Lower triangular factor stored in AB.
  !
  !  N       (input) INTEGER
  !          The order of the matrix A.  N >= 0.
  !
  !  KD      (input) INTEGER
  !          The number of superdiagonals of the matrix A if UPLO = 'U',
  !          or the number of subdiagonals if UPLO = 'L'.  KD >= 0.
  !
  !  NRHS    (input) INTEGER
  !          The number of right hand sides, i.e., the number of columns
  !          of the matrix B.  NRHS >= 0.
  !
  !  AB      (input) DOUBLE PRECISION array, dimension (LDAB,N)
  !          The triangular factor U or L from the Cholesky factorization
  !          A = U**T*U or A = L*L**T of the band matrix A, stored in the
  !          first KD+1 rows of the array.  The j-th column of U or L is
  !          stored in the j-th column of the array AB as follows:
  !          if UPLO ='U', AB(kd+1+i-j,j) = U(i,j) for max(1,j-kd)<=i<=j;
  !          if UPLO ='L', AB(1+i-j,j)    = L(i,j) for j<=i<=min(n,j+kd).
  !
  !  LDAB    (input) INTEGER
  !          The leading dimension of the array AB.  LDAB >= KD+1.
  !
  !  B       (input/output) DOUBLE PRECISION array, dimension (LDB,NRHS)
  !          On entry, the right hand side matrix B.
  !          On exit, the solution matrix X.
  !
  !  LDB     (input) INTEGER
  !          The leading dimension of the array B.  LDB >= max(1,N).
  !
  !  INFO    (output) INTEGER
  !          = 0:  successful exit
  !          < 0:  if INFO = -i, the i-th argument had an illegal value
  !
  ! Call LAPACK routine
  call dpbtrs (uplo, n, kd, nrhs, ab, ldab, b, ldb, info )
  call mth_fail(name,'MTH_DPBTRS',info,error)
end subroutine mth_dpbtrs
! End of Linear Algebra
!
subroutine mth_fail (fac,prog,ifail,error)
  use gkernel_interfaces
  !---------------------------------------------------------------------
  ! Error handling routine
  !---------------------------------------------------------------------
  character(len=*) :: fac           !
  character(len=*) :: prog          !
  integer :: ifail                  !
  logical :: error                  !
  ! Local
  integer :: l1,l2
  character(len=60) :: chain
  !
  if (ifail.eq.0) then
    error = .false.
  else
    l1=lenc(fac)
    l2=lenc(prog)
    write(chain,'(A,A,A,A,A,I4)')   &
     &      'F-',fac(1:l1),',  ERROR in ',prog(1:l2),', ifail = ',ifail
    call gagout(chain)
    error = .true.
  endif
end subroutine mth_fail
!
subroutine do_p_gain(ntimes,nantes,uvsol,refant,error)
  use gildas_def
  use gkernel_interfaces
  use image_def
  use gbl_message
  !
  integer, intent(in) :: ntimes
  integer, intent(in) :: nantes
  character(len=*) :: uvsol
  integer, intent(in) :: refant
  logical, intent(inout) :: error
  !
  character(len=filename_length) :: name
  type(gildas) :: tab
  integer :: iref, nant, ier, i
  real(8) :: time, ztime
  real(4) :: phases(nantes) ! automatic array
  !
  call gildas_null(tab)
  !
  name  = uvsol
  call sic_parsef(name,tab%file,' ','.tab')
  call map_message(seve%i,'UV_GAIN','Creating solution table '//trim(tab%file))
  !
  tab%gil%ndim = 2
  tab%gil%dim(1) = ntimes
  tab%gil%dim(2) = 3+nantes
  ztime = 0
  !
  rewind(2)
  !
  allocate(tab%r2d(tab%gil%dim(1),tab%gil%dim(2)), stat=ier)
  do i=1,ntimes
    read(2,*,iostat=ier) iref, time, nant
    if (ztime.eq.0) ztime = time
    tab%r2d(:,1) = time-ztime
    tab%r2d(:,2) = refant
    tab%r2d(:,3) = nant
    !
    read(2,*) phases(1:nant)
    tab%r2d(i,4:3+nant) = phases(1:nant)-phases(refant)
  enddo
  call gdf_write_image(tab,tab%r2d,error)
  deallocate(tab%r2d)
  close(2)
end subroutine
